# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import posixpath
from typing import Any, Dict, Optional, List

import pyarrow

from deltacat.storage.model.locator import Locator, LocatorName
from deltacat.storage.model.namespace import (
    NamespaceLocator,
    Namespace,
)
from deltacat.storage.model.metafile import Metafile, MetafileRevisionInfo
from deltacat.constants import TXN_DIR_NAME
from deltacat.types.tables import TableProperty

TableProperties = Dict[str, Any]


class Table(Metafile):
    """
    Tables store properties common to every table version including the
    table's name, a high-level description of all table versions, and
    properties shared by all table versions.
    """

    @staticmethod
    def of(
        locator: Optional[TableLocator],
        description: Optional[str] = None,
        properties: Optional[TableProperties] = None,
        latest_active_table_version: Optional[str] = None,
        latest_table_version: Optional[str] = None,
        native_object: Optional[Any] = None,
    ) -> Table:
        table = Table()
        table.locator = locator
        table.description = description
        table.properties = properties
        table.latest_active_table_version = latest_active_table_version
        table.latest_table_version = latest_table_version
        table.native_object = native_object
        return table

    @property
    def locator(self) -> Optional[TableLocator]:
        val: Dict[str, Any] = self.get("tableLocator")
        if val is not None and not isinstance(val, TableLocator):
            self.locator = val = TableLocator(val)
        return val

    @locator.setter
    def locator(self, table_locator: Optional[TableLocator]) -> None:
        self["tableLocator"] = table_locator

    @property
    def description(self) -> Optional[str]:
        return self.get("description")

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self["description"] = description

    @property
    def properties(self) -> Optional[TableProperties]:
        return self.get("properties")

    @properties.setter
    def properties(self, properties: Optional[TableProperties]) -> None:
        self["properties"] = properties

    @property
    def latest_active_table_version(self) -> Optional[str]:
        return self.get("latest_active_table_version")

    @latest_active_table_version.setter
    def latest_active_table_version(
        self,
        latest_active_table_version: Optional[str],
    ) -> None:
        self["latest_active_table_version"] = latest_active_table_version

    @property
    def latest_table_version(self) -> Optional[str]:
        return self.get("latest_table_version")

    @latest_table_version.setter
    def latest_table_version(
        self,
        latest_table_version: Optional[str],
    ) -> None:
        self["latest_table_version"] = latest_table_version

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")

    @native_object.setter
    def native_object(self, native_object: Optional[Any]) -> None:
        self["nativeObject"] = native_object

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        table_locator = self.locator
        if table_locator:
            return table_locator.namespace_locator
        return None

    @property
    def namespace(self) -> Optional[str]:
        table_locator = self.locator
        if table_locator:
            return table_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        table_locator = self.locator
        if table_locator:
            return table_locator.table_name
        return None

    @table_name.setter
    def table_name(self, table_name: Optional[str]) -> None:
        table_locator = self.locator
        if table_locator:
            table_locator.table_name = table_name

    def url(self, catalog_name: Optional[str] = None) -> str:
        return (
            f"dc://{catalog_name}/{self.namespace}/{self.table_name}/"
            if catalog_name
            else f"table://{self.namespace}/{self.table_name}/"
        )

    def read_table_property(self, property: TableProperty) -> Any:
        return TableProperty.read_table_property(self, property)

    def to_serializable(self) -> Table:
        serializable = self
        if serializable.namespace_locator:
            serializable: Table = Table.update_for(self)
            # remove the mutable namespace locator
            serializable.locator.namespace_locator = NamespaceLocator.of(self.id)
        return serializable

    def from_serializable(
        self,
        path: str,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
    ) -> Table:
        # restore the namespace locator from its mapped immutable metafile ID
        if self.namespace_locator and self.namespace_locator.namespace == self.id:
            parent_rev_dir_path = Metafile._parent_metafile_rev_dir_path(
                base_metafile_path=path,
                parent_number=1,
            )
            txn_log_dir = posixpath.join(
                posixpath.dirname(
                    posixpath.dirname(parent_rev_dir_path),
                ),
                TXN_DIR_NAME,
            )
            namespace = Namespace.read(
                MetafileRevisionInfo.latest_revision(
                    revision_dir_path=parent_rev_dir_path,
                    filesystem=filesystem,
                    success_txn_log_dir=txn_log_dir,
                ).path,
                filesystem,
            )
            self.locator.namespace_locator = namespace.locator
        return self


class TableLocatorName(LocatorName):
    def __init__(self, locator: TableLocator):
        self.locator = locator

    @property
    def immutable_id(self) -> Optional[str]:
        return None

    def parts(self) -> List[str]:
        return [self.locator.table_name]


class TableLocator(Locator, dict):
    @staticmethod
    def of(
        namespace_locator: Optional[NamespaceLocator], table_name: Optional[str]
    ) -> TableLocator:
        table_locator = TableLocator()
        table_locator.namespace_locator = namespace_locator
        table_locator.table_name = table_name
        return table_locator

    @staticmethod
    def at(namespace: Optional[str], table_name: Optional[str]) -> TableLocator:
        namespace_locator = NamespaceLocator.of(namespace) if namespace else None
        return TableLocator.of(namespace_locator, table_name)

    @property
    def name(self) -> TableLocatorName:
        return TableLocatorName(self)

    @property
    def parent(self) -> Optional[NamespaceLocator]:
        return self.namespace_locator

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        val: Dict[str, Any] = self.get("namespaceLocator")
        if val is not None and not isinstance(val, NamespaceLocator):
            self.namespace_locator = val = NamespaceLocator(val)
        return val

    @namespace_locator.setter
    def namespace_locator(self, namespace_locator: Optional[NamespaceLocator]) -> None:
        self["namespaceLocator"] = namespace_locator

    @property
    def table_name(self) -> Optional[str]:
        return self.get("tableName")

    @table_name.setter
    def table_name(self, table_name: Optional[str]) -> None:
        self["tableName"] = table_name

    @property
    def namespace(self) -> Optional[str]:
        namespace_locator = self.namespace_locator
        if namespace_locator:
            return namespace_locator.namespace
        return None



import pyarrow as pa
from typing import Callable, Dict, List, Optional, Set, Tuple, Union
from deltacat.tests.compute.test_util_common import (
    offer_iso8601_timestamp_list,
    PartitionKey,
    PartitionKeyType,
    assert_compaction_audit,
    assert_compaction_audit_no_hash_bucket,
)
from deltacat.tests.compute.test_util_constant import (
    DEFAULT_MAX_RECORDS_PER_FILE,
    DEFAULT_HASH_BUCKET_COUNT,
)
from dataclasses import dataclass, fields

from deltacat.compute.compactor.compaction_session import (
    compact_partition_from_request as compact_partition_v1,
)
from deltacat.utils.common import ReadKwargsProvider

from deltacat.storage import (
    DeltaType,
    EntryParams,
)
from deltacat.compute.compactor_v2.compaction_session import (
    compact_partition as compact_partition_v2,
)

from deltacat.compute.compactor.model.compactor_version import CompactorVersion

from deltacat.storage.model.sort_key import SortKey

from deltacat.exceptions import ValidationError

ZERO_VALUED_SORT_KEY, ZERO_VALUED_PARTITION_VALUES_PARAM = [], []
ZERO_VALUED_PARTITION_KEYS_PARAM = None
ZERO_VALUED_PRIMARY_KEY = {}
ZERO_VALUED_PROPERTIES = {}

EMPTY_UTSV_PATH = "deltacat/tests/utils/data/empty.csv"

ENABLED_COMPACT_PARTITIONS_DRIVERS: List[Tuple[CompactorVersion, Callable]] = [
    (CompactorVersion.V1, compact_partition_v1),
    (CompactorVersion.V2, compact_partition_v2),
]


@dataclass(frozen=True)
class BaseCompactorTestCase:
    """
    A pytest parameterized test case for the `compact_partition` function.

    Args:
        primary_keys: Set[str] - argument for the primary_keys parameter in compact_partition. Also needed for table/delta creation
        sort_keys: List[SortKey] - argument for the sort_keys parameter in compact_partition. Also needed for table/delta creation
        partition_keys: List[PartitionKey] - argument for the partition_keys parameter. Needed for table/delta creation
        partition_values: List[Optional[str]] - argument for the partition_valued parameter. Needed for table/delta creation
        input_deltas: List[pa.Array] - argument required for delta creation during compact_partition test setup. Actual incoming deltas expressed as a PyArrow array (https://arrow.apache.org/docs/python/generated/pyarrow.array.html)
        input_deltas_delta_type: DeltaType - enumerated argument required for delta creation during compact_partition test setup. Available values are (DeltaType.APPEND, DeltaType.UPSERT, DeltaType.DELETE). DeltaType.APPEND is not supported by compactor v1 or v2
        expected_terminal_compact_partition_result: pa.Table - expected PyArrow table after compaction (i.e,. the state of the table after applying all row UPDATES/DELETES/INSERTS)
        expected_terminal_exception: BaseException - expected exception during compaction
        expected_terminal_exception_message: Optional[str] - expected exception message if present.
        do_create_placement_group: bool - toggles whether to create a placement group (https://docs.ray.io/en/latest/ray-core/scheduling/placement-group.html) or not
        records_per_compacted_file: int - argument for the records_per_compacted_file parameter in compact_partition
        hash_bucket_count: int - argument for the hash_bucket_count parameter in compact_partition
        read_kwargs_provider: Optional[ReadKwargsProvider] - argument for read_kwargs_provider parameter in compact_partition. If None then no ReadKwargsProvider is provided to compact_partition_params
        drop_duplicates: bool - argument for drop_duplicates parameter in compact_partition. Only recognized by compactor v2.
        skip_enabled_compact_partition_drivers: List[CompactorVersion] - skip whatever enabled_compact_partition_drivers are included in this list
        assert_compaction_audit: Optional[Callable] - argument that asserts compaction_audit is updated only if compactor_version is v2.
    """

    primary_keys: Set[str]
    sort_keys: List[Optional[SortKey]]
    partition_keys: Optional[List[PartitionKey]]
    partition_values: List[Optional[str]]
    input_deltas: Union[List[pa.Array], pa.Table]
    input_deltas_delta_type: DeltaType
    expected_terminal_compact_partition_result: pa.Table
    expected_terminal_exception: BaseException
    expected_terminal_exception_message: str
    do_create_placement_group: bool
    records_per_compacted_file: int
    hash_bucket_count: int
    read_kwargs_provider: Optional[ReadKwargsProvider]
    drop_duplicates: bool
    skip_enabled_compact_partition_drivers: List[CompactorVersion]
    assert_compaction_audit: Optional[Callable]

    # makes CompactorTestCase iterable which is required to build the list of pytest.param values to pass to pytest.mark.parametrize
    def __iter__(self):
        return (getattr(self, field.name) for field in fields(self))


@dataclass(frozen=True)
class IncrementalCompactionTestCaseParams(BaseCompactorTestCase):
    """
    Args:
        is_inplace: bool - argument to indicate whether to try compacting an in-place compacted table (the source table is the destination table). Also needed to control whether the destination table is created
        add_late_deltas: List[Tuple[pa.Table, DeltaType]] - argument to indicate whether to add deltas to the source_partition after we've triggered compaction
    """

    is_inplace: bool
    add_late_deltas: Optional[List[Tuple[pa.Table, DeltaType, Optional[EntryParams]]]]


def with_compactor_version_func_test_param(
    test_cases: Dict[str, BaseCompactorTestCase] = None
):
    test_cases = {} if test_cases is None else test_cases
    enriched_test_cases = {}
    for tc_name, tc_params in test_cases.items():
        for (
            compactor_version,
            compact_partition_func,
        ) in ENABLED_COMPACT_PARTITIONS_DRIVERS:
            # skip creating test case if included in the skip params
            # (e.g. skip_enabled_compact_partition_drivers=[CompactorVersion.V1] will skip creating a compactor version 1 test case)
            if (
                tc_params.skip_enabled_compact_partition_drivers
                and compactor_version
                in tc_params.skip_enabled_compact_partition_drivers
            ):
                continue
            enriched_test_cases[f"{tc_name}_{compactor_version}"] = [
                *tc_params,
                compact_partition_func,
                compactor_version,
            ]
    return enriched_test_cases


INCREMENTAL_TEST_CASES: Dict[str, IncrementalCompactionTestCaseParams] = {
    "1-incremental-pkstr-sknone-norci": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [pa.array([str(i) for i in range(10)])],
            names=["pk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [pa.array([str(i) for i in range(10)])],
            names=["pk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "2-incremental-pkstr-skstr-norci": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array(["test"] * 10),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [pa.array([str(i) for i in range(10)]), pa.array(["test"] * 10)],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "3-incremental-pkstr-multiskstr-norci": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        # column_names=["pk_col_1", "sk_col_1", "sk_col_2"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array(["test"] * 10),
                pa.array(["foo"] * 10),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array(["test"] * 10),
                pa.array(["foo"] * 10),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "4-incremental-duplicate-pk": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key=["sk_col_1"]),
            SortKey.of(key=["sk_col_2"]),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(5)] + ["6", "6", "6", "6", "6"]),
                pa.array([str(i) for i in range(10)]),
                pa.array(["foo"] * 10),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(5)] + ["6"]),
                pa.array([str(i) for i in range(5)] + ["9"]),
                pa.array(["foo"] * 6),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "5-incremental-decimal-pk-simple": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "6-incremental-integer-pk-simple": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "7-incremental-timestamp-pk-simple": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array(offer_iso8601_timestamp_list(10, "minutes")),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(offer_iso8601_timestamp_list(10, "minutes")),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "8-incremental-decimal-timestamp-pk-multi": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1", "pk_col_2"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 20)]),
                pa.array(offer_iso8601_timestamp_list(20, "minutes")),
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 20)]),
                pa.array(offer_iso8601_timestamp_list(20, "minutes")),
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "9-incremental-decimal-pk-multi-dup": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
                pa.array(reversed([i for i in range(20)])),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0.1, 0.2, 0.3, 0.4, 0.5]),
                pa.array([19, 15, 11, 7, 3]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "10-incremental-decimal-pk-partitionless": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=ZERO_VALUED_PARTITION_KEYS_PARAM,
        partition_values=ZERO_VALUED_PARTITION_VALUES_PARAM,
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
                pa.array([i for i in range(20)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0.1, 0.2, 0.3, 0.4, 0.5]),
                pa.array([3, 7, 11, 15, 19]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "11-incremental-decimal-hash-bucket-single": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=ZERO_VALUED_PARTITION_KEYS_PARAM,
        partition_values=ZERO_VALUED_PARTITION_VALUES_PARAM,
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
                pa.array([i for i in range(20)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0.1, 0.2, 0.3, 0.4, 0.5]),
                pa.array([3, 7, 11, 15, 19]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit,
    ),
    "12-incremental-decimal-single-hash-bucket": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=ZERO_VALUED_PARTITION_KEYS_PARAM,
        partition_values=ZERO_VALUED_PARTITION_VALUES_PARAM,
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
                pa.array([i for i in range(20)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0.1, 0.2, 0.3, 0.4, 0.5]),
                pa.array([3, 7, 11, 15, 19]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=1,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=None,
        assert_compaction_audit=assert_compaction_audit_no_hash_bucket,
    ),
    "13-incremental-pkstr-skexists-isinplacecompacted": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=True,
        add_late_deltas=[
            (
                pa.Table.from_arrays(
                    [
                        pa.array([str(i) for i in range(20)]),
                        pa.array([i for i in range(20)]),
                    ],
                    names=["pk_col_1", "sk_col_1"],
                ),
                DeltaType.UPSERT,
                None,
            )
        ],
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=None,
    ),
    "14-incremental-pkstr-skexists-unhappy-hash-bucket-count-not-present": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=ValidationError,
        expected_terminal_exception_message="One of the assertions in DeltaCAT has failed",
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=None,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=False,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=None,
    ),
    "15-incremental-empty-input-with-single-hash-bucket": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key=["sk_col_1"])],
        partition_keys=ZERO_VALUED_PARTITION_KEYS_PARAM,
        partition_values=ZERO_VALUED_PARTITION_VALUES_PARAM,
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([]),
                pa.array([]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([]),
                pa.array([]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        expected_terminal_exception=None,
        expected_terminal_exception_message=None,
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=1,
        read_kwargs_provider=None,
        drop_duplicates=True,
        is_inplace=False,
        add_late_deltas=None,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
        assert_compaction_audit=assert_compaction_audit_no_hash_bucket,
    ),
}

INCREMENTAL_TEST_CASES = with_compactor_version_func_test_param(INCREMENTAL_TEST_CASES)



import typing
from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Protocol,
    Generator,
    Any,
    TypeVar,
    Type,
    Generic,
    List,
    Iterator,
    Optional,
)

from deltacat.experimental.storage.rivulet.fs.file_provider import FileProvider
from deltacat.experimental.storage.rivulet.metastore.sst import SSTableRow
from deltacat.experimental.storage.rivulet.schema.schema import Schema

FILE_FORMAT = TypeVar("FILE_FORMAT")
MEMORY_FORMAT = TypeVar("MEMORY_FORMAT")

T = TypeVar("T")


@dataclass
class RowAndKey(Generic[FILE_FORMAT]):
    """
    Named tuple for a record batch with an index into a specific row
    Note that record batches store data by column, so the row index should be
    used to index into each column array
    """

    row: FILE_FORMAT
    key: Any


class FileReader(
    Protocol[FILE_FORMAT],
    Iterator[RowAndKey[FILE_FORMAT]],
    typing.ContextManager,
):
    """
    Interface for reading specific file

    TODO (IO abstraction) we will need to think about how various IO interfaces (S3, filesystem, memory)
      plug into this.
    """

    @abstractmethod
    def __init__(
        self,
        sst_row: SSTableRow,
        file_provider: FileProvider,
        primary_key: str,
        schema: Schema,
    ) -> None:
        """
        Required constructor (see: FileReaderRegistrar)

        :param sst_row: SSTableRow containing file metadata
        :param file_store: Object providing file access
        """
        ...

    @abstractmethod
    def peek(self) -> Optional[RowAndKey[FILE_FORMAT]]:
        """
        Peek at the next RowAndPrimaryKey without advancing the iterator
        :return: Optional of RowAndPrimaryKey
        """
        ...

    @abstractmethod
    def __next__(self) -> RowAndKey[FILE_FORMAT]:
        """
        Fetch the next RowAndPrimaryKey and advance iterator
        """
        ...

    @abstractmethod
    def close(self):
        """
        Explicit add close so that resources can be cleaned up outside the ContextManager.

        We expect that callers opening the reader can EITHER use a with statement or call __enter__()
        Callers closing the reader can EITHER explicitly call close() or have with statement manage calling __exit__
        """
        ...


class DataReader(Protocol[FILE_FORMAT]):
    """
    Interface for reading specific file formats
    A DatasetReader uses a different DataReader for each format

    TODO (IO abstraction) we will need to think about how various IO interfaces (S3, filesystem, memory)
      plug into this.
    """

    @abstractmethod
    def deserialize_records(
        self, records: FILE_FORMAT, output_type: Type[MEMORY_FORMAT]
    ) -> Generator[MEMORY_FORMAT, None, None]:
        """
        Deserialize records into the specified format.

        Note that output_type gets set based on what a DataScan converts results to,
            e.g. to_arrow, to_dict

        :param records: Input data (generated by generate_records method)
        :param output_type: Type to deserialize into
        :returns: A generator yielding records of the specified type.
        """
        ...

    @abstractmethod
    def join_deserialize_records(
        self,
        records: List[FILE_FORMAT],
        output_type: Type[MEMORY_FORMAT],
        join_key: str,
    ) -> Generator[MEMORY_FORMAT, None, None]:
        """
        Deserialize records into the specified format.

        Note that output_type gets set based on what a DataScan converts results to,
            e.g. to_arrow, to_dict

        :param records: Multiple records which should be merged into final output record
            Note this is a list instead of a set to not enforce hashability
        :param join_key name of field to join across record. This field must be present on all records
        :param output_type: Type to deserialize into
        :returns: A generator yielding records of the specified type.
        """
        ...



import unittest
from unittest.mock import MagicMock, ANY, call
from deltacat.utils.metrics import (
    metrics,
    success_metric,
    failure_metric,
    latency_metric,
    MetricsConfigCache,
    MetricsConfig,
    MetricsTarget,
    METRICS_TARGET_TO_EMITTER_DICT,
)
from deltacat.utils.performance import timed_invocation


def method_not_annotated(mock_func):
    mock_func("called")


@metrics
def metrics_annotated_method(mock_func):
    mock_func("called")


@metrics
def metrics_annotated_method_error(mock_func):
    raise ValueError()


@metrics(prefix="test_prefix")
def metrics_with_prefix_annotated_method(mock_func):
    mock_func("called")


@metrics(prefix="test_prefix")
def metrics_with_prefix_annotated_method_error(mock_func):
    raise ValueError()


class TestMetricsAnnotation(unittest.TestCase):
    def test_metrics_annotation_sanity(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        metrics_annotated_method(mock)

        mock.assert_called_once()
        self.assertEqual(2, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(
                    metrics_name="metrics_annotated_method_time",
                    metrics_config=ANY,
                    value=ANY,
                ),
                call(
                    metrics_name="metrics_annotated_method_success_count",
                    metrics_config=ANY,
                    value=ANY,
                ),
            ],
            any_order=True,
        )

    def test_metrics_annotation_when_error(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        self.assertRaises(ValueError, lambda: metrics_annotated_method_error(mock))

        mock.assert_not_called()
        self.assertEqual(3, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(
                    metrics_name="metrics_annotated_method_error_time",
                    metrics_config=ANY,
                    value=ANY,
                ),
                call(
                    metrics_name="metrics_annotated_method_error_failure_count",
                    metrics_config=ANY,
                    value=ANY,
                ),
                call(
                    metrics_name="metrics_annotated_method_error_failure_count.ValueError",
                    metrics_config=ANY,
                    value=ANY,
                ),
            ],
            any_order=True,
        )

    def test_metrics_with_prefix_annotation_sanity(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        metrics_with_prefix_annotated_method(mock)

        mock.assert_called_once()
        self.assertEqual(2, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(metrics_name="test_prefix_time", metrics_config=ANY, value=ANY),
                call(
                    metrics_name="test_prefix_success_count",
                    metrics_config=ANY,
                    value=ANY,
                ),
            ],
            any_order=True,
        )

    def test_metrics_annotation_performance(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action with annotation
        _, actual_latency = timed_invocation(metrics_annotated_method, mock)
        _, second_call_latency = timed_invocation(metrics_annotated_method, mock)

        mock.assert_called()
        self.assertEqual(4, mock_target.call_count)
        self.assertLess(
            second_call_latency,
            actual_latency,
            "Second call to actor must be much faster",
        )

    def test_metrics_with_prefix_annotation_when_error(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        self.assertRaises(
            ValueError, lambda: metrics_with_prefix_annotated_method_error(mock)
        )

        mock.assert_not_called()
        self.assertEqual(3, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(metrics_name="test_prefix_time", metrics_config=ANY, value=ANY),
                call(
                    metrics_name="test_prefix_failure_count",
                    metrics_config=ANY,
                    value=ANY,
                ),
                call(
                    metrics_name="test_prefix_failure_count.ValueError",
                    metrics_config=ANY,
                    value=ANY,
                ),
            ],
            any_order=True,
        )

    def test_metrics_with_prefix_annotation_without_config(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        MetricsConfigCache.metrics_config = None

        # action
        self.assertRaises(
            ValueError, lambda: metrics_with_prefix_annotated_method_error(mock)
        )

        mock.assert_not_called()
        mock_target.assert_not_called()


@latency_metric
def latency_metric_annotated_method(mock_func):
    mock_func("called")


@latency_metric
def latency_metric_annotated_method_error(mock_func):
    raise ValueError()


@latency_metric(name="test")
def latency_metric_with_name_annotated_method(mock_func):
    mock_func("called")


@latency_metric(name="test")
def latency_metric_with_name_annotated_method_error(mock_func):
    raise ValueError()


class TestLatencyMetricAnnotation(unittest.TestCase):
    def test_annotation_sanity(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        latency_metric_annotated_method(mock)

        mock.assert_called_once()
        self.assertEqual(1, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(
                    metrics_name="latency_metric_annotated_method_time",
                    metrics_config=ANY,
                    value=ANY,
                )
            ],
            any_order=True,
        )

    def test_annotation_when_error(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        self.assertRaises(
            ValueError, lambda: latency_metric_annotated_method_error(mock)
        )

        mock.assert_not_called()
        self.assertEqual(1, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(
                    metrics_name="latency_metric_annotated_method_error_time",
                    metrics_config=ANY,
                    value=ANY,
                )
            ],
            any_order=True,
        )

    def test_annotation_with_args_sanity(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        latency_metric_with_name_annotated_method(mock)

        mock.assert_called_once()
        self.assertEqual(1, mock_target.call_count)
        mock_target.assert_has_calls(
            [call(metrics_name="test", metrics_config=ANY, value=ANY)], any_order=True
        )

    def test_annotation_with_args_when_error(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        self.assertRaises(
            ValueError, lambda: latency_metric_with_name_annotated_method_error(mock)
        )

        mock.assert_not_called()
        self.assertEqual(1, mock_target.call_count)
        mock_target.assert_has_calls(
            [call(metrics_name="test", metrics_config=ANY, value=ANY)], any_order=True
        )

    def test_annotation_without_config(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        MetricsConfigCache.metrics_config = None

        # action
        self.assertRaises(
            ValueError, lambda: latency_metric_with_name_annotated_method_error(mock)
        )

        mock.assert_not_called()
        mock_target.assert_not_called()


@success_metric
def success_metric_annotated_method(mock_func):
    mock_func("called")


@success_metric
def success_metric_annotated_method_error(mock_func):
    raise ValueError()


@success_metric(name="test")
def success_metric_with_name_annotated_method(mock_func):
    mock_func("called")


@success_metric(name="test")
def success_metric_with_name_annotated_method_error(mock_func):
    raise ValueError()


class TestSuccessMetricAnnotation(unittest.TestCase):
    def test_annotation_sanity(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        success_metric_annotated_method(mock)

        mock.assert_called_once()
        self.assertEqual(1, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(
                    metrics_name="success_metric_annotated_method_success_count",
                    metrics_config=ANY,
                    value=ANY,
                )
            ],
            any_order=True,
        )

    def test_annotation_when_error(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        self.assertRaises(
            ValueError, lambda: success_metric_annotated_method_error(mock)
        )

        mock.assert_not_called()
        mock_target.assert_not_called()

    def test_annotation_with_args_sanity(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        success_metric_with_name_annotated_method(mock)

        mock.assert_called_once()
        self.assertEqual(1, mock_target.call_count)
        mock_target.assert_has_calls(
            [call(metrics_name="test", metrics_config=ANY, value=ANY)], any_order=True
        )

    def test_annotation_with_args_when_error(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        self.assertRaises(
            ValueError, lambda: success_metric_with_name_annotated_method_error(mock)
        )

        mock.assert_not_called()
        mock_target.assert_not_called()

    def test_annotation_without_config(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        MetricsConfigCache.metrics_config = None

        # action
        success_metric_annotated_method(mock)

        mock.assert_called_once()
        mock_target.assert_not_called()


@failure_metric
def failure_metric_annotated_method(mock_func):
    mock_func("called")


@failure_metric
def failure_metric_annotated_method_error(mock_func):
    raise ValueError()


@failure_metric(name="test")
def failure_metric_with_name_annotated_method(mock_func):
    mock_func("called")


@failure_metric(name="test")
def failure_metric_with_name_annotated_method_error(mock_func):
    raise ValueError()


class TestFailureMetricAnnotation(unittest.TestCase):
    def test_annotation_sanity(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        failure_metric_annotated_method(mock)

        mock.assert_called_once()
        mock_target.assert_not_called()

    def test_annotation_when_error(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        self.assertRaises(
            ValueError, lambda: failure_metric_annotated_method_error(mock)
        )

        mock.assert_not_called()
        self.assertEqual(2, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(
                    metrics_name="failure_metric_annotated_method_error_failure_count.ValueError",
                    metrics_config=ANY,
                    value=ANY,
                ),
                call(
                    metrics_name="failure_metric_annotated_method_error_failure_count",
                    metrics_config=ANY,
                    value=ANY,
                ),
            ],
            any_order=True,
        )

    def test_annotation_with_args_sanity(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        failure_metric_with_name_annotated_method(mock)

        mock.assert_called_once()
        mock_target.assert_not_called()

    def test_annotation_with_args_when_error(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        config = MetricsConfig("us-east-1", MetricsTarget.NOOP)
        MetricsConfigCache.metrics_config = config

        # action
        self.assertRaises(
            ValueError, lambda: failure_metric_with_name_annotated_method_error(mock)
        )

        mock.assert_not_called()
        self.assertEqual(2, mock_target.call_count)
        mock_target.assert_has_calls(
            [
                call(metrics_name="test.ValueError", metrics_config=ANY, value=ANY),
                call(metrics_name="test", metrics_config=ANY, value=ANY),
            ],
            any_order=True,
        )

    def test_annotation_without_config(self):
        mock, mock_target = MagicMock(), MagicMock()
        METRICS_TARGET_TO_EMITTER_DICT[MetricsTarget.NOOP] = mock_target
        MetricsConfigCache.metrics_config = None

        # action
        self.assertRaises(
            ValueError, lambda: failure_metric_with_name_annotated_method_error(mock)
        )

        mock.assert_not_called()
        mock_target.assert_not_called()



from __future__ import annotations
from enum import Enum
from typing import Callable, Optional, TYPE_CHECKING
import logging

import tenacity

from pyarrow.lib import ArrowException, ArrowInvalid, ArrowCapacityError

import botocore
from botocore.exceptions import BotoCoreError

import ray
from ray.exceptions import (
    RayError,
    RayTaskError,
    RuntimeEnvSetupError,
    WorkerCrashedError,
    NodeDiedError,
    OutOfMemoryError,
)

from daft.exceptions import DaftTransientError, DaftCoreException

import deltacat as dc
from deltacat import logs
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
)

if TYPE_CHECKING:
    from deltacat.storage.model.schema import FieldLocator

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

DELTACAT_STORAGE_PARAM = "deltacat_storage"
DELTACAT_STORAGE_KWARGS_PARAM = "deltacat_storage_kwargs"


class DeltaCatErrorNames(str, Enum):

    DEPENDENCY_RAY_ERROR = "DependencyRayError"
    DEPENDENCY_RAY_WORKER_DIED_ERROR = "DependencyRayWorkerDiedError"
    DEPENDENCY_RAY_OUT_OF_MEMORY_ERROR = "DependencyRayOOMError"
    DEPENDENCY_RAY_RUNTIME_SETUP_ERROR = "DependencyRayRuntimeSetupError"
    DEPENDENCY_BOTOCORE_ERROR = "DependencyBotocoreError"
    DEPENDENCY_BOTOCORE_CONNECTION_ERROR = "DependencyBotocoreConnectionError"
    DEPENDENCY_BOTOCORE_CREDENTIAL_ERROR = "DependencyBotocoreCredentialError"
    DEPENDENCY_BOTOCORE_TIMEOUT_ERROR = "DependencyBotocoreTimeoutError"
    NON_RETRYABLE_DOWNLOAD_TABLE_ERROR = "NonRetryableDownloadTableError"
    NON_RETRYABLE_DOWNLOAD_FILE_ERROR = "NonRetryableDownloadFileError"
    NON_RETRYABLE_UPLOAD_TABLE_ERROR = "NonRetryableUploadTableError"
    NON_RETRYABLE_UPLOAD_FILE_ERROR = "NonRetryableUploadFileError"
    DEPENDENCY_PYARROW_ERROR = "DependencyPyarrowError"
    DEPENDENCY_PYARROW_INVALID_ERROR = "DependencyPyarrowInvalidError"
    DEPENDENCY_PYARROW_CAPACITY_ERROR = "DependencyPyarrowCapacityError"
    PYMEMCACHED_PUT_OBJECT_ERROR = "PymemcachedPutObjectError"
    DEPENDENCY_DAFT_ERROR = "DependencyDaftError"

    GENERAL_THROTTLING_ERROR = "GeneralThrottlingError"
    RETRYABLE_UPLOAD_TABLE_ERROR = "RetryableUploadTableError"
    RETRYABLE_UPLOAD_FILE_ERROR = "RetryableUploadFileError"
    RETRYABLE_DOWNLOAD_FILE_ERROR = "RetryableDownloadFileError"
    RETRYABLE_DOWNLOAD_TABLE_ERROR = "RetryableDownloadTableError"
    RETRYABLE_TIMEOUT_ERROR = "RetryableTimeoutError"
    DEPENDENCY_DAFT_TRANSIENT_ERROR = "DependencyDaftTransientError"

    VALIDATION_ERROR = "ValidationError"
    CONTENT_TYPE_VALIDATION_ERROR = "ContentTypeValidationError"

    DELTACAT_SYSTEM_ERROR = "DeltaCatSystemError"
    DELTACAT_TRANSIENT_ERROR = "DeltaCatTransientError"
    UNCLASSIFIED_DELTACAT_ERROR = "UnclassifiedDeltaCatError"
    UNRECOGNIZED_RAY_TASK_ERROR = "UnrecognizedRayTaskError"

    NAMESPACE_NOT_FOUND_ERROR = "NamespaceNotFoundError"
    TABLE_NOT_FOUND_ERROR = "TableNotFoundError"
    TABLE_VERSION_NOT_FOUND_ERROR = "TableVersionNotFoundError"
    STREAM_NOT_FOUND_ERROR = "StreamNotFoundError"
    PARTITION_NOT_FOUND_ERROR = "PartitionNotFoundError"
    DELTA_NOT_FOUND_ERROR = "DeltaNotFoundError"
    TABLE_ALREADY_EXISTS_ERROR = "TableAlreadyExistsError"
    TABLE_VERSION_ALREADY_EXISTS_ERROR = "TableVersionAlreadyExistsError"
    NAMESPACE_ALREADY_EXISTS_ERROR = "NamespaceAlreadyExistsError"
    SCHEMA_COMPATIBILITY_ERROR = "SchemaCompatibilityError"
    SCHEMA_VALIDATION_ERROR = "SchemaValidationError"
    TABLE_VALIDATION_ERROR = "TableValidationError"
    CONCURRENT_MODIFICATION_ERROR = "ConcurrentModificationError"
    OBJECT_NOT_FOUND_ERROR = "ObjectNotFoundError"
    OBJECT_DELETED_ERROR = "ObjectDeletedError"
    OBJECT_ALREADY_EXISTS_ERROR = "ObjectAlreadyExistsError"


class DeltaCatError(Exception):
    def __init__(self, *args, **kwargs):
        task_id, node_ip = self._get_ray_task_id_and_node_ip()
        self.task_id = task_id
        self.node_ip = node_ip
        super().__init__(*args, **kwargs)

    def _get_ray_task_id_and_node_ip(self):
        if ray.is_initialized():
            task_id = get_current_ray_task_id()
            node_ip = ray.util.get_node_ip_address()
            return task_id, node_ip
        else:
            return None, None


class NonRetryableError(DeltaCatError):
    is_retryable = False


class RetryableError(DeltaCatError):
    is_retryable = True


class ValidationError(NonRetryableError):
    error_name = DeltaCatErrorNames.VALIDATION_ERROR.value


class UnclassifiedDeltaCatError(NonRetryableError):
    error_name = DeltaCatErrorNames.UNCLASSIFIED_DELTACAT_ERROR.value


class DependencyRayError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_RAY_ERROR.value


class DeltaCatTransientError(RetryableError):
    error_name = DeltaCatErrorNames.DELTACAT_TRANSIENT_ERROR.value


class DependencyDaftError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_DAFT_ERROR.value


class DependencyRayWorkerDiedError(RetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_RAY_WORKER_DIED_ERROR.value


class DependencyRayOutOfMemoryError(RetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_RAY_OUT_OF_MEMORY_ERROR.value


class DependencyRayRuntimeSetupError(RetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_RAY_RUNTIME_SETUP_ERROR.value


class DependencyPyarrowError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_PYARROW_ERROR.value


class DependencyPyarrowInvalidError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_PYARROW_INVALID_ERROR.value


class DependencyPyarrowCapacityError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_PYARROW_CAPACITY_ERROR.value


class PymemcachedPutObjectError(RetryableError):
    error_name = DeltaCatErrorNames.PYMEMCACHED_PUT_OBJECT_ERROR.value


class ContentTypeValidationError(NonRetryableError):
    error_name = DeltaCatErrorNames.CONTENT_TYPE_VALIDATION_ERROR.value


class DependencyBotocoreError(NonRetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_BOTOCORE_ERROR.value


class DependencyBotocoreConnectionError(DeltaCatTransientError):
    error_name = DeltaCatErrorNames.DEPENDENCY_BOTOCORE_CONNECTION_ERROR.value


class DependencyBotocoreCredentialError(DeltaCatTransientError):
    error_name = DeltaCatErrorNames.DEPENDENCY_BOTOCORE_CREDENTIAL_ERROR.value


class DependencyBotocoreTimeoutError(DeltaCatTransientError):
    error_name = DeltaCatErrorNames.DEPENDENCY_BOTOCORE_TIMEOUT_ERROR.value


class NonRetryableDownloadFileError(NonRetryableError):
    error_name = DeltaCatErrorNames.NON_RETRYABLE_DOWNLOAD_FILE_ERROR.value


class NonRetryableDownloadTableError(NonRetryableDownloadFileError):
    error_name = DeltaCatErrorNames.NON_RETRYABLE_DOWNLOAD_TABLE_ERROR.value


class NonRetryableUploadFileError(NonRetryableError):
    error_name = DeltaCatErrorNames.NON_RETRYABLE_UPLOAD_FILE_ERROR.value


class NonRetryableUploadTableError(NonRetryableUploadFileError):
    error_name = DeltaCatErrorNames.NON_RETRYABLE_UPLOAD_TABLE_ERROR.value


class GeneralThrottlingError(RetryableError):
    error_name = DeltaCatErrorNames.GENERAL_THROTTLING_ERROR.value


class RetryableUploadFileError(RetryableError):
    error_name = DeltaCatErrorNames.RETRYABLE_UPLOAD_FILE_ERROR.value


class RetryableUploadTableError(RetryableUploadFileError):
    error_name = DeltaCatErrorNames.RETRYABLE_UPLOAD_TABLE_ERROR.value


class RetryableDownloadFileError(RetryableError):
    error_name = DeltaCatErrorNames.RETRYABLE_DOWNLOAD_FILE_ERROR.value


class RetryableDownloadTableError(RetryableDownloadFileError):
    error_name = DeltaCatErrorNames.RETRYABLE_DOWNLOAD_TABLE_ERROR.value


class RetryableTimeoutError(RetryableError):
    error_name = DeltaCatErrorNames.RETRYABLE_TIMEOUT_ERROR.value


class DependencyDaftTransientError(RetryableError):
    error_name = DeltaCatErrorNames.DEPENDENCY_DAFT_TRANSIENT_ERROR.value


class DeltaCatSystemError(NonRetryableError):
    error_name = DeltaCatErrorNames.DELTACAT_SYSTEM_ERROR.value


class UnrecognizedRayTaskError(NonRetryableError):
    error_name = DeltaCatErrorNames.UNRECOGNIZED_RAY_TASK_ERROR.value


class NamespaceNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.NAMESPACE_NOT_FOUND_ERROR.value


class TableNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_NOT_FOUND_ERROR.value


class TableVersionNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_VERSION_NOT_FOUND_ERROR.value


class PartitionNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.PARTITION_NOT_FOUND_ERROR.value


class StreamNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.STREAM_NOT_FOUND_ERROR.value


class DeltaNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.DELTA_NOT_FOUND_ERROR.value


class TableAlreadyExistsError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_ALREADY_EXISTS_ERROR.value


class TableVersionAlreadyExistsError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_VERSION_ALREADY_EXISTS_ERROR.value


class NamespaceAlreadyExistsError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_ALREADY_EXISTS_ERROR.value


class ObjectNotFoundError(NonRetryableError):
    error_name = DeltaCatErrorNames.OBJECT_NOT_FOUND_ERROR.value


class ObjectDeletedError(NonRetryableError):
    error_name = DeltaCatErrorNames.OBJECT_DELETED_ERROR.value


class ObjectAlreadyExistsError(NonRetryableError):
    error_name = DeltaCatErrorNames.OBJECT_ALREADY_EXISTS_ERROR.value


class ConcurrentModificationError(NonRetryableError):
    error_name = DeltaCatErrorNames.CONCURRENT_MODIFICATION_ERROR.value


class SchemaValidationError(NonRetryableError):
    error_name = DeltaCatErrorNames.SCHEMA_VALIDATION_ERROR.value


class TableValidationError(NonRetryableError):
    error_name = DeltaCatErrorNames.TABLE_VALIDATION_ERROR.value


class SchemaCompatibilityError(NonRetryableError):
    error_name = DeltaCatErrorNames.SCHEMA_COMPATIBILITY_ERROR.value
    """Raised when a schema update would break backward compatibility."""

    def __init__(
        self,
        message: str,
        field_locator: Optional[FieldLocator] = None,
        *args,
        **kwargs,
    ):
        super().__init__(message, *args, **kwargs)
        self.field_locator = field_locator


def categorize_errors(func: Callable):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            deltacat_storage = None
            deltacat_storage_kwargs = {}
            all_args = args
            if kwargs:
                deltacat_storage = kwargs.get(DELTACAT_STORAGE_PARAM)
                deltacat_storage_kwargs = kwargs.get(DELTACAT_STORAGE_KWARGS_PARAM, {})
                all_args = all_args + tuple(kwargs.values())

            if not deltacat_storage and all_args:
                for arg in all_args:
                    if (
                        isinstance(arg, dict)
                        and arg.get(DELTACAT_STORAGE_PARAM) is not None
                    ):
                        deltacat_storage = arg.get(DELTACAT_STORAGE_PARAM)
                        deltacat_storage_kwargs = arg.get(
                            DELTACAT_STORAGE_KWARGS_PARAM, {}
                        )
                        break

            categorize_deltacat_exception(e, deltacat_storage, deltacat_storage_kwargs)

    return wrapper


def categorize_deltacat_exception(
    e: BaseException,
    deltacat_storage: dc.storage.interface = None,
    deltacat_storage_kwargs: dict = None,
):
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}

    if isinstance(e, DeltaCatError):
        raise e
    elif deltacat_storage and deltacat_storage.can_categorize(
        e, **deltacat_storage_kwargs
    ):
        deltacat_storage.raise_categorized_error(e, **deltacat_storage_kwargs)
    elif isinstance(e, RayError):
        _categorize_ray_error(e)
    elif isinstance(e, tenacity.RetryError):
        _categorize_tenacity_error(e)
    elif isinstance(e, ArrowException):
        _categorize_dependency_pyarrow_error(e)
    elif isinstance(e, AssertionError):
        _categorize_assertion_error(e)
    elif isinstance(e, DaftCoreException):
        _categorize_daft_error(e)
    elif isinstance(e, BotoCoreError):
        _categorize_botocore_error(e)
    else:
        _categorize_all_remaining_errors(e)

    logger.error(f"Error categorization failed for {e}.", exc_info=True)
    raise UnclassifiedDeltaCatError(
        "Error could not categorized into DeltaCat error"
    ) from e


def _categorize_ray_error(e: RayError):
    if isinstance(e, RuntimeEnvSetupError):
        raise DependencyRayRuntimeSetupError("Ray failed to setup runtime env.") from e
    elif isinstance(e, WorkerCrashedError) or isinstance(e, NodeDiedError):
        raise DependencyRayWorkerDiedError("Ray worker died unexpectedly.") from e
    elif isinstance(e, OutOfMemoryError):
        raise DependencyRayOutOfMemoryError("Ray worker Out Of Memory.") from e
    elif isinstance(e, RayTaskError):
        if e.cause is not None and isinstance(e.cause, Exception):
            categorize_deltacat_exception(e.cause)
        else:
            raise UnrecognizedRayTaskError(
                "Unrecognized underlying error detected in a Ray task."
            ) from e
    else:
        raise DependencyRayError("Dependency Ray error occurred.") from e


def _categorize_tenacity_error(e: tenacity.RetryError):
    if e.__cause__ is not None and isinstance(e.__cause__, Exception):
        categorize_deltacat_exception(e.__cause__)
    else:
        raise RetryableError("Unrecognized retryable error occurred.") from e


def _categorize_dependency_pyarrow_error(e: ArrowException):
    if isinstance(e, ArrowInvalid):
        raise DependencyPyarrowInvalidError(
            f"Pyarrow Invalid error occurred. {e}"
        ) from e
    elif isinstance(e, ArrowCapacityError):
        raise DependencyPyarrowCapacityError("Pyarrow Capacity error occurred.") from e
    else:
        raise DependencyPyarrowError("Pyarrow error occurred.") from e


def _categorize_assertion_error(e: BaseException):
    raise ValidationError(f"One of the assertions in DeltaCAT has failed. {e}") from e


def _categorize_daft_error(e: DaftCoreException):
    if isinstance(e, DaftTransientError):
        raise DependencyDaftTransientError("Daft Transient error occurred.") from e
    elif isinstance(e, DaftCoreException):
        raise DependencyDaftError("Daft error occurred.") from e


def _categorize_botocore_error(e: BotoCoreError):
    if isinstance(e, botocore.exceptions.ConnectionError) or isinstance(
        e, botocore.exceptions.HTTPClientError
    ):
        raise DependencyBotocoreConnectionError(
            "Botocore connection error occurred."
        ) from e
    elif isinstance(e, botocore.exceptions.CredentialRetrievalError) or isinstance(
        e, botocore.exceptions.NoCredentialsError
    ):
        raise DependencyBotocoreCredentialError(
            "Botocore credential retrieval failed"
        ) from e
    elif isinstance(e, botocore.exceptions.ReadTimeoutError) or isinstance(
        e, botocore.exceptions.ConnectTimeoutError
    ):
        raise DependencyBotocoreTimeoutError("Botocore connection timed out.") from e
    else:
        raise DependencyBotocoreError("Botocore error occurred.") from e


def _categorize_all_remaining_errors(e: BaseException):
    if isinstance(e, ConnectionError):
        raise DeltaCatTransientError("Connection error has occurred.") from e
    elif isinstance(e, TimeoutError):
        raise DeltaCatTransientError("Timeout error has occurred.") from e
    elif isinstance(e, OSError):
        raise DeltaCatTransientError("OSError occurred.") from e
    elif isinstance(e, SystemExit):
        raise DeltaCatSystemError("Unexpected System error occurred.") from e



import shutil
import tempfile


from contextlib import contextmanager


@contextmanager
def temp_dir_autocleanup():
    tmpdir = tempfile.mkdtemp()
    try:
        yield tmpdir
    finally:
        shutil.rmtree(tmpdir)



from dataclasses import dataclass

from deltacat.storage.model.scan.scan_task import ScanTask


@dataclass
class ScanPlan:
    """Represents collection of ScanTasks to be passed to compute engine for query planning"""

    scan_tasks: list[ScanTask]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import json
import posixpath

import pyarrow

from typing import Any, Dict, List, Optional, TYPE_CHECKING

from deltacat.storage.model.metafile import Metafile, MetafileRevisionInfo
from deltacat.constants import TXN_DIR_NAME
from deltacat.storage.model.schema import (
    FieldLocator,
)
from deltacat.storage.model.locator import (
    Locator,
    LocatorName,
)
from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.storage.model.stream import StreamLocator
from deltacat.storage.model.table import (
    TableLocator,
    Table,
)
from deltacat.storage.model.table_version import TableVersionLocator
from deltacat.storage.model.transform import Transform
from deltacat.storage.model.types import (
    CommitState,
    StreamFormat,
)
from deltacat.types.media import ContentType

if TYPE_CHECKING:
    from deltacat.compute.compactor import RoundCompletionInfo


"""
An ordered list of partition values. Partition values are typically derived
by applying one or more transforms to a table's fields.
"""
PartitionValues = List[Any]

"""
Constants for special partition types.
"""
UNPARTITIONED_SCHEME_NAME = "unpartitioned_scheme"
UNPARTITIONED_SCHEME_ID = "deadbeef-7277-49a4-a195-fdc8ed235d42"
UNKNOWN_PARTITION_ID = "deadbeef-2fe7-4557-82c9-da53b1862003"  # a partition ID that is assumed to exist but is not known
UNSPECIFIED_PARTITION_ID = "deadbeef-5bff-41ea-b82c-e531f445632b"  # a partition ID that has been left intentionally unspecified


class Partition(Metafile):
    @staticmethod
    def of(
        locator: Optional[PartitionLocator],
        content_types: Optional[List[ContentType]],
        state: Optional[CommitState] = None,
        previous_stream_position: Optional[int] = None,
        previous_partition_id: Optional[str] = None,
        stream_position: Optional[int] = None,
        partition_scheme_id: Optional[str] = None,
        compaction_round_completion_info: Optional[RoundCompletionInfo] = None,
    ) -> Partition:
        partition = Partition()
        partition.locator = locator
        partition.content_types = content_types
        partition.state = state
        partition.previous_stream_position = previous_stream_position
        partition.previous_partition_id = previous_partition_id
        partition.stream_position = stream_position
        partition.partition_scheme_id = (
            partition_scheme_id
            if locator and locator.partition_values
            else UNPARTITIONED_SCHEME_ID
        )
        partition.compaction_round_completion_info = compaction_round_completion_info
        return partition

    @property
    def locator(self) -> Optional[PartitionLocator]:
        val: Dict[str, Any] = self.get("partitionLocator")
        if val is not None and not isinstance(val, PartitionLocator):
            self.locator = val = PartitionLocator(val)
        return val

    @locator.setter
    def locator(self, partition_locator: Optional[PartitionLocator]) -> None:
        self["partitionLocator"] = partition_locator

    @property
    def locator_alias(self) -> Optional[PartitionLocatorAlias]:
        return PartitionLocatorAlias.of(self)

    @property
    def content_types(self) -> Optional[List[ContentType]]:
        content_types = self.get("contentTypes")
        return (
            None
            if content_types is None
            else [None if _ is None else ContentType(_) for _ in content_types]
        )

    @content_types.setter
    def content_types(
        self, supported_content_types: Optional[List[ContentType]]
    ) -> None:
        self["contentTypes"] = supported_content_types

    @property
    def state(self) -> Optional[CommitState]:
        state = self.get("state")
        return None if state is None else CommitState(state)

    @state.setter
    def state(self, state: Optional[CommitState]) -> None:
        self["state"] = state

    @property
    def previous_stream_position(self) -> Optional[int]:
        return self.get("previousStreamPosition")

    @previous_stream_position.setter
    def previous_stream_position(self, previous_stream_position: Optional[int]) -> None:
        self["previousStreamPosition"] = previous_stream_position

    @property
    def previous_partition_id(self) -> Optional[str]:
        return self.get("previousPartitionId")

    @previous_partition_id.setter
    def previous_partition_id(self, previous_partition_id: Optional[str]) -> None:
        self["previousPartitionId"] = previous_partition_id

    @property
    def stream_position(self) -> Optional[int]:
        return self.get("streamPosition")

    @stream_position.setter
    def stream_position(self, stream_position: Optional[int]):
        self["streamPosition"] = stream_position

    @property
    def partition_scheme_id(self) -> Optional[str]:
        return self.get("partitionSchemeId")

    @partition_scheme_id.setter
    def partition_scheme_id(self, partition_scheme_id: Optional[str]) -> None:
        self["partitionSchemeId"] = partition_scheme_id

    @property
    def compaction_round_completion_info(self) -> Optional[RoundCompletionInfo]:
        """
        Round completion info for compaction operations.
        This replaces the need for separate round completion files.
        """
        val: Dict[str, Any] = self.get("compactionRoundCompletionInfo")
        if val is not None:
            # Import here to avoid circular imports
            from deltacat.compute.compactor import RoundCompletionInfo

            if not isinstance(val, RoundCompletionInfo):
                self["compactionRoundCompletionInfo"] = val = RoundCompletionInfo(val)
        return val

    @compaction_round_completion_info.setter
    def compaction_round_completion_info(
        self, compaction_round_completion_info: Optional[RoundCompletionInfo]
    ) -> None:
        self["compactionRoundCompletionInfo"] = compaction_round_completion_info

    @property
    def partition_id(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.partition_id
        return None

    @property
    def stream_id(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.stream_id
        return None

    @property
    def stream_format(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.stream_format
        return None

    @property
    def partition_values(self) -> Optional[PartitionValues]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.partition_values
        return None

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_version_locator
        return None

    @property
    def stream_locator(self) -> Optional[StreamLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.stream_locator
        return None

    @property
    def storage_type(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.stream_format
        return None

    @property
    def namespace(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_version
        return None

    def url(self, catalog_name: Optional[str] = None) -> str:
        return (
            f"dc://{catalog_name}/{self.namespace}/{self.table_name}/{self.table_version}/{self.stream_format}/{json.dumps(self.partition_values)}/"
            if catalog_name
            else f"table://{self.namespace}/{self.table_name}/{self.table_version}/{self.stream_format}/{json.dumps(self.partition_values)}/"
        )

    def is_supported_content_type(self, content_type: ContentType) -> bool:
        supported_content_types = self.content_types
        return (not supported_content_types) or (
            content_type in supported_content_types
        )

    def to_serializable(self) -> Partition:
        serializable: Partition = Partition.update_for(self)
        if serializable.table_locator:
            # replace the mutable table locator
            serializable.table_version_locator.table_locator = TableLocator.at(
                namespace=self.id,
                table_name=self.id,
            )
        return serializable

    def from_serializable(
        self,
        path: str,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
    ) -> Partition:
        # restore the table locator from its mapped immutable metafile ID
        if self.table_locator and self.table_locator.table_name == self.id:
            parent_rev_dir_path = Metafile._parent_metafile_rev_dir_path(
                base_metafile_path=path,
                parent_number=3,
            )
            txn_log_dir = posixpath.join(
                posixpath.dirname(
                    posixpath.dirname(
                        posixpath.dirname(parent_rev_dir_path),
                    )
                ),
                TXN_DIR_NAME,
            )
            table = Table.read(
                MetafileRevisionInfo.latest_revision(
                    revision_dir_path=parent_rev_dir_path,
                    filesystem=filesystem,
                    success_txn_log_dir=txn_log_dir,
                ).path,
                filesystem,
            )
            self.table_version_locator.table_locator = table.locator
        return self


class PartitionLocatorName(LocatorName):
    def __init__(self, locator: PartitionLocator):
        self.locator = locator

    @property
    def immutable_id(self) -> Optional[str]:
        return self.locator.partition_id

    @immutable_id.setter
    def immutable_id(self, immutable_id: Optional[str]):
        self.locator.partition_id = immutable_id

    def parts(self) -> List[str]:
        return [
            str(self.locator.partition_values),
            self.locator.partition_id,
        ]


class PartitionLocator(Locator, dict):
    @staticmethod
    def of(
        stream_locator: Optional[StreamLocator],
        partition_values: Optional[PartitionValues],
        partition_id: Optional[str],
    ) -> PartitionLocator:
        """
        Creates a stream partition locator. Partition ID is
        case-sensitive.

        Partition Value types must ensure that
        `str(partition_value1) == str(partition_value2)` always implies
        `partition_value1 == partition_value2` (i.e. if two string
        representations of partition values are equal, then the two partition
        values are equal).
        """
        partition_locator = PartitionLocator()
        partition_locator.stream_locator = stream_locator
        partition_locator.partition_values = partition_values
        partition_locator.partition_id = partition_id
        return partition_locator

    @staticmethod
    def at(
        namespace: Optional[str],
        table_name: Optional[str],
        table_version: Optional[str],
        stream_id: Optional[str],
        stream_format: Optional[StreamFormat],
        partition_values: Optional[PartitionValues],
        partition_id: Optional[str],
    ) -> PartitionLocator:
        stream_locator = (
            StreamLocator.at(
                namespace,
                table_name,
                table_version,
                stream_id,
                stream_format,
            )
            if stream_format or stream_id
            else None
        )
        return PartitionLocator.of(
            stream_locator,
            partition_values,
            partition_id,
        )

    @property
    def name(self) -> PartitionLocatorName:
        return PartitionLocatorName(self)

    @property
    def parent(self) -> Optional[StreamLocator]:
        return self.stream_locator

    @property
    def stream_locator(self) -> Optional[StreamLocator]:
        val: Dict[str, Any] = self.get("streamLocator")
        if val is not None and not isinstance(val, StreamLocator):
            self.stream_locator = val = StreamLocator(val)
        return val

    @stream_locator.setter
    def stream_locator(self, stream_locator: Optional[StreamLocator]) -> None:
        self["streamLocator"] = stream_locator

    @property
    def partition_values(self) -> Optional[PartitionValues]:
        return self.get("partitionValues")

    @partition_values.setter
    def partition_values(self, partition_values: Optional[PartitionValues]) -> None:
        self["partitionValues"] = (
            partition_values or None
        )  # normalize empty partition values to None

    @property
    def partition_id(self) -> Optional[str]:
        return self.get("partitionId")

    @partition_id.setter
    def partition_id(self, partition_id: Optional[str]) -> None:
        self["partitionId"] = partition_id

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_version_locator
        return None

    @property
    def stream_id(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.stream_id
        return None

    @property
    def stream_format(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.format
        return None

    @property
    def namespace(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_version
        return None


class PartitionKey(dict):
    @staticmethod
    def of(
        key: List[FieldLocator],
        name: Optional[str] = None,
        field_id: Optional[int] = None,
        transform: Optional[Transform] = None,
        native_object: Optional[Any] = None,
    ) -> PartitionKey:
        if (
            len(key) > 1
            and transform is not None
            and not transform.is_multi_field_transform
        ):
            raise ValueError(f"{len(key)} keys given for 1-key transform.")
        return PartitionKey(
            {
                "key": key,
                "name": name,
                "fieldId": field_id,
                "transform": transform,
                "nativeObject": native_object,
            }
        )

    def equivalent_to(
        self,
        other: PartitionKey,
        check_identifiers: False,
    ):
        if other is None:
            return False
        if not isinstance(other, dict):
            return False
        if not isinstance(other, PartitionKey):
            other = PartitionKey(other)
        return (
            self.key == other.key
            and self.transform == other.transform
            and not check_identifiers
            or (self.name == other.name and self.id == other.id)
        )

    @property
    def key(self) -> List[FieldLocator]:
        return self.get("key")

    @property
    def name(self) -> Optional[str]:
        return self.get("name")

    @property
    def id(self) -> Optional[int]:
        return self.get("fieldId")

    @property
    def transform(self) -> Optional[Transform]:
        val: Dict[str, Any] = self.get("transform")
        if val is not None and not isinstance(val, Transform):
            self["transform"] = val = Transform(val)
        return val

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")


class PartitionKeyList(List[PartitionKey]):
    @staticmethod
    def of(items: List[PartitionKey]) -> PartitionKeyList:
        typed_items = PartitionKeyList()
        for item in items:
            if item is not None and not isinstance(item, PartitionKey):
                item = PartitionKey(item)
            typed_items.append(item)
        return typed_items

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, PartitionKey):
            self[item] = val = PartitionKey(val)
        return val

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]  # This triggers __getitem__ conversion


class PartitionScheme(dict):
    @staticmethod
    def of(
        keys: Optional[PartitionKeyList],
        name: Optional[str] = None,
        scheme_id: Optional[str] = None,
        native_object: Optional[Any] = None,
    ) -> PartitionScheme:
        # Validate keys if provided
        if keys is not None:
            # Check for empty keys list
            if len(keys) == 0:
                raise ValueError("Partition scheme cannot have empty keys list")

            # Check for duplicate keys (by field locators and transform types) and names
            seen_key_transform_pairs = set()
            seen_names = set()
            for key in keys:
                # Check for duplicate field locators with identical transform types
                key_tuple = tuple(key.key) if key.key else ()
                transform_type = type(key.transform) if key.transform else None
                key_transform_pair = (key_tuple, transform_type)

                if key_transform_pair in seen_key_transform_pairs:
                    # Use the first field locator for the error message
                    key_name = key.key[0] if key.key else "unknown"
                    transform_name = (
                        transform_type.__name__ if transform_type else "None"
                    )
                    raise ValueError(
                        f"Duplicate partition key found: {key_name} with transform type {transform_name}"
                    )
                seen_key_transform_pairs.add(key_transform_pair)

                # Check for duplicate names (when specified)
                if key.name is not None:
                    if key.name in seen_names:
                        raise ValueError(
                            f"Duplicate partition key name found: {key.name}"
                        )
                    seen_names.add(key.name)

        return PartitionScheme(
            {
                "keys": keys,
                "name": name,
                "id": scheme_id,
                "nativeObject": native_object,
            }
        )

    def equivalent_to(
        self,
        other: PartitionScheme,
        check_identifiers: bool = False,
    ) -> bool:
        if other is None:
            return False
        if not isinstance(other, dict):
            return False
        if not isinstance(other, PartitionScheme):
            other = PartitionScheme(other)
        # If both have None keys, they are equivalent (for unpartitioned schemes)
        if self.keys is None and other.keys is None:
            return not check_identifiers or (
                self.name == other.name and self.id == other.id
            )
        # If only one has None keys, they are not equivalent
        if self.keys is None or other.keys is None:
            return False
        # Compare keys if both have them
        for i in range(len(self.keys)):
            if not self.keys[i].equivalent_to(other.keys[i], check_identifiers):
                return False
        return not check_identifiers or (
            self.name == other.name and self.id == other.id
        )

    @property
    def keys(self) -> Optional[PartitionKeyList]:
        val: List[PartitionKey] = self.get("keys")
        if val is not None and not isinstance(val, PartitionKeyList):
            self["keys"] = val = PartitionKeyList.of(val)
        return val

    @property
    def name(self) -> Optional[str]:
        return self.get("name")

    @property
    def id(self) -> Optional[str]:
        return self.get("id")

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")


UNPARTITIONED_SCHEME = PartitionScheme.of(
    keys=None,
    name=UNPARTITIONED_SCHEME_NAME,
    scheme_id=UNPARTITIONED_SCHEME_ID,
)


class PartitionSchemeList(List[PartitionScheme]):
    @staticmethod
    def of(items: List[PartitionScheme]) -> PartitionSchemeList:
        typed_items = PartitionSchemeList()
        for item in items:
            if item is not None and not isinstance(item, PartitionScheme):
                item = PartitionScheme(item)
            typed_items.append(item)
        return typed_items

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, PartitionScheme):
            self[item] = val = PartitionScheme(val)
        return val

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]  # This triggers __getitem__ conversion


class PartitionLocatorAliasName(LocatorName):
    def __init__(self, locator: PartitionLocatorAlias):
        self.locator = locator

    @property
    def immutable_id(self) -> Optional[str]:
        return None

    def parts(self) -> List[str]:
        return [
            str(self.locator.partition_values),
            self.locator.partition_scheme_id,
        ]


class PartitionLocatorAlias(Locator, dict):
    @staticmethod
    def of(parent_partition: Partition):
        return (
            PartitionLocatorAlias(
                {
                    "partition_values": parent_partition.partition_values,
                    "partition_scheme_id": parent_partition.partition_scheme_id,
                    "parent": (
                        parent_partition.locator.parent
                        if parent_partition.locator
                        else None
                    ),
                }
            )
            if parent_partition.state != CommitState.STAGED
            else None  # staged partitions cannot be resolved by alias
        )

    @property
    def partition_values(self) -> Optional[PartitionValues]:
        return self.get("partition_values")

    @property
    def partition_scheme_id(self) -> Optional[str]:
        return self.get("partition_scheme_id")

    @property
    def name(self) -> PartitionLocatorAliasName:
        return PartitionLocatorAliasName(self)

    @property
    def parent(self) -> Optional[Locator]:
        return self.get("parent")



import logging
from typing import Any, Dict, Generator, Optional
from botocore.config import Config
from deltacat.aws.constants import (
    BOTO_MAX_RETRIES,
    BOTO_THROTTLING_ERROR_CODES,
)
from deltacat.constants import (
    UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY,
    RETRYABLE_TRANSIENT_ERRORS,
)

from boto3.resources.base import ServiceResource
from botocore.client import BaseClient
from botocore.exceptions import ClientError
from tenacity import (
    Retrying,
    retry_if_exception_type,
    stop_after_delay,
    wait_random_exponential,
)
import deltacat.aws.clients as aws_utils
from deltacat import logs
from deltacat.exceptions import (
    RetryableError,
    RetryableDownloadFileError,
    RetryableUploadFileError,
    NonRetryableDownloadFileError,
    NonRetryableUploadFileError,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class S3Url:
    def __init__(self, url: str):

        from urllib.parse import urlparse

        self._parsed = urlparse(url, allow_fragments=False)  # support '#' in path
        if not self._parsed.scheme:  # support paths w/o 's3://' scheme
            url = f"s3://{url}"
            self._parsed = urlparse(url, allow_fragments=False)
        if self._parsed.query:  # support '?' in path
            self.key = f"{self._parsed.path.lstrip('/')}?{self._parsed.query}"
        else:
            self.key = self._parsed.path.lstrip("/")
        self.bucket = self._parsed.netloc
        self.url = self._parsed.geturl()


def parse_s3_url(url: str) -> S3Url:
    return S3Url(url)


def s3_resource_cache(region: Optional[str], **kwargs) -> ServiceResource:

    return aws_utils.resource_cache(
        "s3",
        region,
        **kwargs,
    )


def s3_client_cache(region: Optional[str], **kwargs) -> BaseClient:

    return aws_utils.client_cache("s3", region, **kwargs)


def get_object_at_url(url: str, **s3_client_kwargs) -> Dict[str, Any]:

    s3 = s3_client_cache(None, **s3_client_kwargs)

    parsed_s3_url = parse_s3_url(url)
    return s3.get_object(Bucket=parsed_s3_url.bucket, Key=parsed_s3_url.key)


def delete_files_by_prefix(bucket: str, prefix: str, **s3_client_kwargs) -> None:

    s3 = s3_resource_cache(None, **s3_client_kwargs)
    bucket = s3.Bucket(bucket)
    bucket.objects.filter(Prefix=prefix).delete()


def filter_paths_by_prefix(bucket, prefix):
    return objects_to_paths(
        bucket,
        filter_objects_by_prefix(bucket, prefix),
    )


def objects_to_paths(bucket, objects):
    for obj in objects:
        yield get_path_from_object(bucket, obj)


def get_path_from_object(bucket, obj):
    return "s3://{}/{}".format(bucket, obj["Key"])


def filter_objects_by_prefix(
    bucket: str, prefix: str, **s3_client_kwargs
) -> Generator[Dict[str, Any], None, None]:

    s3 = s3_client_cache(None, **s3_client_kwargs)
    params = {"Bucket": bucket, "Prefix": prefix}
    more_objects_to_list = True
    while more_objects_to_list:
        response = s3.list_objects_v2(**params)
        if "Contents" in response:
            for obj in response["Contents"]:
                yield obj
        params["ContinuationToken"] = response.get("NextContinuationToken")
        more_objects_to_list = params["ContinuationToken"] is not None


def upload(s3_url: str, body, **s3_client_kwargs) -> Dict[str, Any]:

    parsed_s3_url = parse_s3_url(s3_url)
    s3 = s3_client_cache(None, **s3_client_kwargs)
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=15),
        stop=stop_after_delay(UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )
    return retrying(
        _put_object,
        s3,
        body,
        parsed_s3_url.bucket,
        parsed_s3_url.key,
    )


def _put_object(
    s3_client, body: Any, bucket: str, key: str, **s3_put_object_kwargs
) -> Dict[str, Any]:
    try:
        return s3_client.put_object(
            Body=body, Bucket=bucket, Key=key, **s3_put_object_kwargs
        )
    except ClientError as e:
        if e.response["Error"]["Code"] in BOTO_THROTTLING_ERROR_CODES:
            error_code = e.response["Error"]["Code"]
            raise RetryableUploadFileError(
                f"Retry upload for: {bucket}/{key} after receiving {error_code}",
            ) from e
        raise NonRetryableUploadFileError(
            f"Failed table upload to: {bucket}/{key}"
        ) from e
    except RETRYABLE_TRANSIENT_ERRORS as e:
        raise RetryableUploadFileError(
            f"Retry upload for: {bucket}/{key} after receiving {type(e).__name__}"
        ) from e
    except BaseException as e:
        logger.error(
            f"Upload has failed for {bucket}/{key}. Error: {type(e).__name__}",
            exc_info=True,
        )
        raise NonRetryableUploadFileError(
            f"Failed table upload to: {bucket}/{key}"
        ) from e


def download(
    s3_url: str, fail_if_not_found: bool = True, **s3_client_kwargs
) -> Optional[Dict[str, Any]]:

    parsed_s3_url = parse_s3_url(s3_url)
    s3 = s3_client_cache(None, **s3_client_kwargs)
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=15),
        stop=stop_after_delay(UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )
    return retrying(
        _get_object,
        s3,
        parsed_s3_url.bucket,
        parsed_s3_url.key,
        fail_if_not_found=fail_if_not_found,
    )


def _get_object(s3_client, bucket: str, key: str, fail_if_not_found: bool = True):
    try:
        return s3_client.get_object(
            Bucket=bucket,
            Key=key,
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "NoSuchKey":
            if fail_if_not_found:
                raise NonRetryableDownloadFileError(
                    f"Failed get object from: {bucket}/{key}"
                ) from e
            logger.info(f"file not found: {bucket}/{key}")
    except RETRYABLE_TRANSIENT_ERRORS as e:
        raise RetryableDownloadFileError(
            f"Retry get object: {bucket}/{key} after receiving {type(e).__name__}"
        ) from e

    return None


def _get_s3_client_kwargs_from_token(token_holder) -> Dict[Any, Any]:
    conf = Config(retries={"max_attempts": BOTO_MAX_RETRIES, "mode": "adaptive"})
    return (
        {
            "aws_access_key_id": token_holder["accessKeyId"],
            "aws_secret_access_key": token_holder["secretAccessKey"],
            "aws_session_token": token_holder["sessionToken"],
            "config": conf,
        }
        if token_holder
        else {"config": conf}
    )



from typing import Generator, Dict, Optional

import pyarrow as pa

from deltacat.storage.model.shard import Shard
from deltacat.experimental.storage.rivulet.reader.dataset_reader import DatasetReader
from deltacat.experimental.storage.rivulet.reader.query_expression import (
    QueryExpression,
)
from deltacat.experimental.storage.rivulet import Schema


class DataScan:
    """
    Top level class representing and executing a data scan, on both riv internal and external data
    This class is lazy, and executed when the user calls a method "to_{format}"
    to deserialize data into the chosen in-memory format

    Dataset.py scan() is the entrypoint to create and return data scan. The user
    then has to chain a "to_{format}" method to read rows in their chosen in-memory format

    Rivulet cannot simply return file URIs and allow query engine to process files,
    because rivulet will internally manage details like indexes, custom file formats for bulk records, where data is physically laid out across row groups, etc.

    DataScan allows query engines to send push down predicates. Push down predicates are used to filter on dimensions natively indexed by riv (e.g. primary key), and also

    DataScan is coupled to internals of the riv rivulet format. If the rivulet format evolves, DataScan execution hould be able to understand which rivulet spec version is used and be compatible with any valid rivule rivulet.

    FUTURE IMPROVEMENTS
    1. Implement full spec for push down predicates
    2. Figure out how permissions/credential providers work.
    3. Figure out how extension libraries can plug in to_x deserialization support.
        One potential option is to override __getattr__ and check a static class-level Registry
        of to_x methods. Modules would have to import DataScan and call DataScan.register_deserializer(...)
    """

    def __init__(
        self,
        dataset_schema: Schema,
        query: QueryExpression,
        dataset_reader: DatasetReader,
        shard: Optional[Shard],
    ):
        self.dataset_schema = dataset_schema
        self.query = query
        self.dataset_reader = dataset_reader
        self.shard = shard

    def to_arrow(self) -> Generator[pa.RecordBatch, None, None]:
        """
        Generates scan results as arrow record batches

        TODO how to make the .to_x methods pluggable?
        """
        return self.dataset_reader.scan(
            self.dataset_schema, pa.RecordBatch, self.query, shard=self.shard
        )

    def to_pydict(self) -> Generator[Dict, None, None]:
        """
        Generates scan results as a Dict for each row
        """
        return self.dataset_reader.scan(
            self.dataset_schema, Dict, self.query, shard=self.shard
        )



import pytest
import os
import pyarrow
import msgpack
import posixpath


from deltacat.storage import (
    Transaction,
    TransactionOperation,
    TransactionOperationType,
    Namespace,
    NamespaceLocator,
    Metafile,
)

from deltacat.constants import (
    TXN_DIR_NAME,
    RUNNING_TXN_DIR_NAME,
    PAUSED_TXN_DIR_NAME,
)


class TestAbsToRelative:
    @classmethod
    def setup_method(cls):
        cls.catalog_root = "/catalog/root/path"

    # Test cases for the abs_to_relative function
    def test_abs_to_relative_simple(self):
        """
        Tests the function which relativizes absolute paths (string) into relative paths (string)
        """
        catalog_root = TestAbsToRelative.catalog_root
        absolute_path = "/catalog/root/path/namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        relative_path = Transaction._abs_txn_meta_path_to_relative(
            catalog_root, absolute_path
        )
        assert (
            relative_path
            == "namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        )

    def test_abs_to_relative_same_paths(self):
        catalog_root = TestAbsToRelative.catalog_root
        absolute_path = TestAbsToRelative.catalog_root
        with pytest.raises(
            ValueError,
            match="Target and root are identical, but expected target to be a child of root.",
        ):
            Transaction._abs_txn_meta_path_to_relative(catalog_root, absolute_path)

    def test_abs_to_relative_root_with_trailing_slash(self):
        catalog_root = "/catalog/root/path/"
        absolute_path = "/catalog/root/path/namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        relative_path = Transaction._abs_txn_meta_path_to_relative(
            catalog_root, absolute_path
        )
        assert (
            relative_path
            == "namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        )

    def test_abs_to_relative_bad_root(self):
        catalog_root = TestAbsToRelative.catalog_root
        absolute_path = "/cat/rt/pth/namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk"
        with pytest.raises(ValueError, match="Expected target to be a child of root."):
            Transaction._abs_txn_meta_path_to_relative(catalog_root, absolute_path)

    def test_abs_to_relative_empty_path(self):
        with pytest.raises(ValueError, match="Expected target to be a child of root."):
            Transaction._abs_txn_meta_path_to_relative("", "/lorem/ipsum")
        with pytest.raises(ValueError, match="Expected target to be a child of root."):
            Transaction._abs_txn_meta_path_to_relative("/lorem/ipsum/", "")

    # Test cases for the relativize_operation_paths function
    def test_relativizemetafile_write_paths(self):
        catalog_root = "/catalog/root"
        absolute_paths = [
            "/catalog/root/path/to/metafile1.mpk",
            "/catalog/root/path/to/metafile2.mpk",
            "/catalog/root/another/path/lore_ipsum.mpk",
            "/catalog/root/another/path/meta/to/lorem_ipsum.mpk",
            "/catalog/root/another/path/lorem_ipsum.mpk",
            "/catalog/root/here.mpk",
        ]
        expected_relative_paths = [
            "path/to/metafile1.mpk",
            "path/to/metafile2.mpk",
            "another/path/lore_ipsum.mpk",
            "another/path/meta/to/lorem_ipsum.mpk",
            "another/path/lorem_ipsum.mpk",
            "here.mpk",
        ]
        # Create a dummy transaction operation with absolute paths
        dest_metafile = Metafile({"id": "dummy_metafile_id"})
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=dest_metafile,
        )
        # use replace method as setter
        transaction_operation.metafile_write_paths = absolute_paths
        # Create a transaction and relativize paths
        transaction = Transaction.of([transaction_operation])
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        # Verify the paths have been correctly relativized
        assert transaction_operation.metafile_write_paths == expected_relative_paths

    def test_relativize_locator_write_paths(self):
        catalog_root = "/catalog/root"
        absolute_paths = [
            "/catalog/root/path/to/loc1.mpk",
            "/catalog/root/path/to/loc2.mpk",
            "/catalog/root/another/path/lore_ipsum.mpk",
            "/catalog/root/another/path/meta/to/lorem_ipsum.mpk",
            "/catalog/root/another/path/lorem_ipsum.mpk",
            "/catalog/root/here.mpk",
        ]
        expected_relative_paths = [
            "path/to/loc1.mpk",
            "path/to/loc2.mpk",
            "another/path/lore_ipsum.mpk",
            "another/path/meta/to/lorem_ipsum.mpk",
            "another/path/lorem_ipsum.mpk",
            "here.mpk",
        ]
        # Create a dummy transaction operation with absolute paths
        dest_metafile = Metafile({"id": "dummy_metafile_id"})
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=dest_metafile,
        )
        # use replace as setter
        transaction_operation.locator_write_paths = absolute_paths
        # Create a transaction and relativize paths
        transaction = Transaction.of(txn_operations=[transaction_operation])
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        # Verify the paths have been correctly relativized
        assert transaction_operation.locator_write_paths == expected_relative_paths

    def test_relativize_metafile_and_locator_paths(self):
        catalog_root = "/meta_catalog/root_dir/a/b/c"
        meta_absolute_paths = [
            "/meta_catalog/root_dir/a/b/c/namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk",
            "/meta_catalog/root_dir/a/b/c/namespace/table/table_version/stream_id/partition_id/00000000000000000002.mpk",
            "/meta_catalog/root_dir/a/b/c/namespace/table/table_version/stream_id/partition_id/00000000000000000003.mpk",
        ]
        loc_absolute_paths = [
            "/meta_catalog/root_dir/a/b/c/d/table/table_version/stream_id/partition_id/00000000000000000001.mpk",
            "/meta_catalog/root_dir/a/b/c/e/table/table_version/stream_id/partition_id/00000000000000000002.mpk",
            "/meta_catalog/root_dir/a/b/c/f/table/table_version/stream_id/partition_id/00000000000000000003.mpk",
        ]
        meta_relative_paths = [
            "namespace/table/table_version/stream_id/partition_id/00000000000000000001.mpk",
            "namespace/table/table_version/stream_id/partition_id/00000000000000000002.mpk",
            "namespace/table/table_version/stream_id/partition_id/00000000000000000003.mpk",
        ]
        loc_relative_paths = [
            "d/table/table_version/stream_id/partition_id/00000000000000000001.mpk",
            "e/table/table_version/stream_id/partition_id/00000000000000000002.mpk",
            "f/table/table_version/stream_id/partition_id/00000000000000000003.mpk",
        ]
        # Create a dummy transaction operation with absolute paths
        dest_metafile = Metafile({"id": "dummy_metafile_id"})
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=dest_metafile,
        )
        # use replace as setter
        transaction_operation.metafile_write_paths = meta_absolute_paths
        transaction_operation.locator_write_paths = loc_absolute_paths
        # Create a transaction and relativize paths
        transaction = Transaction.of([transaction_operation])
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        # Verify the paths have been correctly relativized
        assert (
            transaction_operation.metafile_write_paths == meta_relative_paths
        ), f"Expected: {meta_relative_paths}, but got: {transaction_operation.metafile_write_paths}"
        assert (
            transaction_operation.locator_write_paths == loc_relative_paths
        ), f"Expected: {loc_relative_paths}, but got: {transaction_operation.locator_write_paths}"

    def test_multiple_operations_relativize_paths(self):
        catalog_root = "/catalog/root"
        meta_absolute_paths = [
            "/catalog/root/path/to/metafile1.mpk",
            "/catalog/root/path/to/metafile2.mpk",
            "/catalog/root/another/path/lore_ipsum.mpk",
            "/catalog/root/another/path/meta/to/lorem_ipsum.mpk",
            "/catalog/root/another/path/lorem_ipsum.mpk",
            "/catalog/root/here.mpk",
        ]
        loc_absolute_paths = [
            "/catalog/root/path/to/loc1.mpk",
            "/catalog/root/path/to/loc2.mpk",
            "/catalog/root/another/path/lore_ipsum.mpk",
            "/catalog/root/another/path/meta/to/lorem_ipsum.mpk",
            "/catalog/root/another/path/lorem_ipsum.mpk",
            "/catalog/root/here.mpk",
        ]
        meta_expected_relative_paths = [
            "path/to/metafile1.mpk",
            "path/to/metafile2.mpk",
            "another/path/lore_ipsum.mpk",
            "another/path/meta/to/lorem_ipsum.mpk",
            "another/path/lorem_ipsum.mpk",
            "here.mpk",
        ]
        loc_expected_relative_paths = [
            "path/to/loc1.mpk",
            "path/to/loc2.mpk",
            "another/path/lore_ipsum.mpk",
            "another/path/meta/to/lorem_ipsum.mpk",
            "another/path/lorem_ipsum.mpk",
            "here.mpk",
        ]
        # Create a dummy transaction operation with absolute paths
        dest_metafile = Metafile({"id": "dummy_metafile_id"})
        transaction_operations = []
        for i in range(11):
            transaction_operation = TransactionOperation.of(
                operation_type=TransactionOperationType.CREATE,
                dest_metafile=dest_metafile,
            )
            transaction_operation.metafile_write_paths = meta_absolute_paths
            transaction_operation.locator_write_paths = loc_absolute_paths
            transaction_operations.append(transaction_operation)
        # Create a transaction and relativize paths
        transaction = Transaction.of(transaction_operations)
        for operation in transaction_operations:
            transaction.relativize_operation_paths(operation, catalog_root)
        # Verify the paths have been correctly relativized
        for operation in transaction_operations:
            assert operation.metafile_write_paths == meta_expected_relative_paths
            assert operation.locator_write_paths == loc_expected_relative_paths

    def test_empty_metafile_and_locator_write_paths(self):
        catalog_root = "/catalog/root"
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=Metafile({"id": "dummy_metafile_id"}),
        )
        # Empty paths
        transaction_operation.metafile_write_paths = []
        transaction_operation.locator_write_paths = []
        transaction = Transaction.of([transaction_operation])
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        assert transaction_operation.metafile_write_paths == []
        assert transaction_operation.locator_write_paths == []

    def test_large_number_of_paths(self):
        catalog_root = "/catalog/root"
        absolute_paths = [f"/catalog/root/path/to/file{i}.mpk" for i in range(5000)]
        expected_paths = [f"path/to/file{i}.mpk" for i in range(5000)]
        transaction_operation = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=Metafile({"id": "dummy_metafile_id"}),
        )
        transaction_operation.metafile_write_paths = absolute_paths
        transaction = Transaction.of([transaction_operation])
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        assert transaction_operation.metafile_write_paths == expected_paths

    def test_large_number_of_paths_multi_ops(self):
        catalog_root = "/catalog/root"
        absolute_paths = [f"/catalog/root/path/to/file{i}.mpk" for i in range(1000)]
        expected_paths = [f"path/to/file{i}.mpk" for i in range(1000)]

        # Different operation types to test
        operation_types = [
            TransactionOperationType.CREATE,
            # TransactionOperationType.UPDATE,
            TransactionOperationType.DELETE,
            TransactionOperationType.READ_EXISTS,
            TransactionOperationType.READ_LATEST,
            TransactionOperationType.READ_CHILDREN,
            TransactionOperationType.READ_SIBLINGS,
        ]

        transaction_ops = []
        for op_type in operation_types:
            transaction_operation = TransactionOperation.of(
                operation_type=op_type,
                dest_metafile=Metafile({"id": "dummy_metafile_id"}),
            )
            transaction_operation.metafile_write_paths = absolute_paths
            transaction_ops.append(transaction_operation)
        transaction = Transaction.of([transaction_operation])
        transaction.relativize_operation_paths(transaction_operation, catalog_root)
        # Assert paths are relativized correctly
        assert (
            transaction_operation.metafile_write_paths == expected_paths
        ), f"Failed for operation type {op_type}"


class TestTransactionPersistence:

    # Verifies that transactions initialized with empty or None operations are marked interactive,
    # while valid operations are not
    def test_create_iterative_transaction(self):
        txn_1 = Transaction.of(txn_operations=[])
        txn_2 = Transaction.of(txn_operations=None)
        op = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=Metafile({"id": "dummy_metafile_id"}),
        )
        txn_3 = Transaction.of(txn_operations=[op, op])
        assert (
            txn_1.interactive
        )  # check if constructor detect empty list --> interactive transaction
        assert (
            txn_2.interactive
        )  # check if we can initialize with no list --> interactive transaction
        assert (
            not txn_3.interactive
        )  # check that valid operations_list --> not interactive transaction

    # Builds and commits a transaction step-by-step, then validates the output files and transaction success log
    def test_commit_iterative_transaction(self, temp_dir):
        # Create two simple namespaces
        namespace_locator1 = NamespaceLocator.of(namespace="test_ns_1")
        namespace_locator2 = NamespaceLocator.of(namespace="test_ns_2")
        ns1 = Namespace.of(locator=namespace_locator1)
        ns2 = Namespace.of(locator=namespace_locator2)
        # Start with an empty transaction (interactive)
        transaction = Transaction.of()
        txn = transaction.start(temp_dir)  # operate on deep-copy
        # Build operations manually and step them in
        op1 = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=ns1,
        )
        op2 = TransactionOperation.of(
            operation_type=TransactionOperationType.CREATE,
            dest_metafile=ns2,
        )
        # steps
        txn.step(op1)
        txn.step(op2)

        # seal() for interactive transactions
        write_paths, success_log_path = txn.seal()

        # Check output files exist and are valid
        deserialized_ns1 = Namespace.read(write_paths[0])
        deserialized_ns2 = Namespace.read(write_paths[1])

        assert ns1.equivalent_to(deserialized_ns1)
        assert ns2.equivalent_to(deserialized_ns2)
        assert success_log_path.endswith(str(txn.end_time))

    # Ensures that stepping and committing a transaction writes non-empty output files and a valid success log
    def test_commit_iterative_file_creation(self, temp_dir):
        ns = Namespace.of(locator=NamespaceLocator.of(namespace="check_writes"))
        txn = Transaction.of().start(temp_dir)
        op = TransactionOperation.of(TransactionOperationType.CREATE, dest_metafile=ns)
        txn.step(op)
        write_paths, success_log_path = txn.seal()

        # check the files were created
        for path in write_paths:
            abs_path = os.path.join(temp_dir, path)
            assert os.path.exists(abs_path)
            assert os.path.getsize(abs_path) > 0

        # check the success log exists
        assert os.path.exists(success_log_path)
        assert os.path.getsize(success_log_path) > 0

    # Confirms that a transaction can be paused, resumed, and successfully committed without data los
    def test_transaction_pause_and_resume_roundtrip(self, temp_dir):
        # Create a test namespace
        ns = Namespace.of(locator=NamespaceLocator.of(namespace="paused_resume_ns"))

        # Start interactive transaction
        txn = Transaction.of().start(temp_dir)
        op = TransactionOperation.of(TransactionOperationType.CREATE, dest_metafile=ns)

        txn.step(op)

        # Pause transaction (writes to paused/)
        txn.pause()

        # Resume transaction (reads from paused/)
        txn.resume()

        # Commit resumed transaction
        write_paths, success_log_path = txn.seal()

        # Validate outputs
        deserialized = Namespace.read(write_paths[0])
        assert ns.equivalent_to(deserialized)
        assert os.path.exists(success_log_path)
        assert success_log_path.endswith(str(txn.end_time))

    # Validates that transaction state, including ID and write paths, is correctly preserved across pause/resume cycles
    def test_resume_preserves_state_after_pause(self, temp_dir):
        ns = Namespace.of(locator=NamespaceLocator.of(namespace="resume_state_check"))

        txn = Transaction.of().start(temp_dir)
        op = TransactionOperation.of(TransactionOperationType.CREATE, dest_metafile=ns)

        txn.step(op)
        txn_id_before = txn.id

        txn.pause()
        txn.resume()

        # Ensure the ID and provider are still valid
        assert txn.id == txn_id_before
        assert txn._time_provider is not None
        assert hasattr(txn, "metafile_write_paths")
        assert len(txn.metafile_write_paths) == 1

        # Check commit still works
        _, success_log_path = txn.seal()
        assert os.path.exists(success_log_path)

    # Explicitly checks that fields are preserved
    def test_resume_preserves_state_after_pause_deep(self, temp_dir):
        ns = Namespace.of(locator=NamespaceLocator.of(namespace="resume_state_check"))

        txn = Transaction.of().start(temp_dir)
        op = TransactionOperation.of(TransactionOperationType.CREATE, dest_metafile=ns)

        txn.step(op)

        # Save values before pause
        txn_id_before = txn.id
        start_time_before = txn.start_time
        root_before = txn.catalog_root_normalized
        meta_paths_before = list(txn.metafile_write_paths)
        locator_paths_before = list(txn.locator_write_paths)

        txn.pause()
        txn.resume()

        # Field-by-field checks
        assert txn.id == txn_id_before, "Transaction ID should be preserved"
        assert txn._time_provider is not None, "Time provider should be reinitialized"
        assert txn.start_time == start_time_before, "Start time should be preserved"
        assert txn.catalog_root_normalized == root_before, "Catalog root should match"
        assert (
            txn.metafile_write_paths == meta_paths_before
        ), "Metafile paths must match"
        assert (
            txn.locator_write_paths == locator_paths_before
        ), "Locator paths must match"
        assert (
            isinstance(txn.operations, list) and len(txn.operations) == 1
        ), "Operations must be restored"
        assert txn.pause_time is not None, "Pause time should be restored"

        # Final commit still works
        write_paths, success_log_path = txn.seal()
        assert os.path.exists(success_log_path)

    # Checks that pausing a transaction moves its log from running/ to paused/ and preserves valid transaction state
    def test_pause_moves_running_to_paused(self, temp_dir):
        # Set up a transaction and a single operation
        locator = NamespaceLocator.of(namespace="pause_test")
        ns = Namespace.of(locator=locator)
        txn = Transaction.of().start(temp_dir)

        op = TransactionOperation.of(TransactionOperationType.CREATE, dest_metafile=ns)
        txn.step(op)

        fs = pyarrow.fs.LocalFileSystem()
        txn_id = txn.id
        txn_log_dir = posixpath.join(temp_dir, TXN_DIR_NAME)

        running_path = posixpath.join(txn_log_dir, RUNNING_TXN_DIR_NAME, txn_id)
        paused_path = posixpath.join(txn_log_dir, PAUSED_TXN_DIR_NAME, txn_id)

        # Sanity check: file should be in running/
        assert fs.get_file_info(running_path).type == pyarrow.fs.FileType.File

        # Pause transaction
        txn.pause()
        # Ensure the running file is deleted
        assert fs.get_file_info(running_path).type == pyarrow.fs.FileType.NotFound

        # Ensure the paused file exists and contains valid msgpack
        paused_info = fs.get_file_info(paused_path)
        assert paused_info.type == pyarrow.fs.FileType.File
        with fs.open_input_stream(paused_path) as f:
            data = f.readall()
            txn_loaded = msgpack.loads(data)
            assert "operations" in txn_loaded

    # Simulates a full multi-step transaction with multiple pause/resume cycles and verifies correctness of all outputs
    def test_transaction_pause_and_resume_roundtrip_complex(self, temp_dir):
        # Step 0: Create an empty interactive transaction
        txn = Transaction.of().start(temp_dir)

        # Step 1: Add first namespace, pause
        ns1 = Namespace.of(locator=NamespaceLocator.of(namespace="roundtrip_ns_1"))
        op1 = TransactionOperation.of(
            TransactionOperationType.CREATE, dest_metafile=ns1
        )
        txn.step(op1)
        txn.pause()

        # Step 2: Resume, add second namespace, pause
        txn.resume()
        ns2 = Namespace.of(locator=NamespaceLocator.of(namespace="roundtrip_ns_2"))
        op2 = TransactionOperation.of(
            TransactionOperationType.CREATE, dest_metafile=ns2
        )
        txn.step(op2)
        txn.pause()

        # Step 3: Resume again, add third namespace, commit
        txn.resume()
        ns3 = Namespace.of(locator=NamespaceLocator.of(namespace="roundtrip_ns_3"))
        op3 = TransactionOperation.of(
            TransactionOperationType.CREATE, dest_metafile=ns3
        )
        txn.step(op3)

        # Final commit
        write_paths, success_log_path = txn.seal()

        # Read and verify written namespaces
        for i, ns in enumerate([ns1, ns2, ns3]):
            written_path = write_paths[i]
            deserialized_ns = Namespace.read(written_path)
            assert ns.equivalent_to(
                deserialized_ns
            ), f"Mismatch in ns{i+1}: {ns} != {deserialized_ns}"
            assert os.path.exists(written_path), f"Missing file: {written_path}"
            assert os.path.getsize(written_path) > 0

        # Check success log exists and is correct
        assert os.path.exists(success_log_path)
        assert success_log_path.endswith(str(txn.end_time))

    # Repeats a complex pause/resume flow with additional assertions on namespace equality and time consistency
    def test_transaction_pause_and_resume_roundtrip_complex_2(self, temp_dir):
        # Step 0: Create an empty interactive transaction
        txn = Transaction.of().start(temp_dir)

        # Step 1: Add first namespace, pause
        ns1 = Namespace.of(locator=NamespaceLocator.of(namespace="roundtrip_ns_1"))
        op1 = TransactionOperation.of(
            TransactionOperationType.CREATE, dest_metafile=ns1
        )
        txn.step(op1)
        txn.pause()

        # Step 2: Resume, add second namespace, pause
        txn.resume()
        ns2 = Namespace.of(locator=NamespaceLocator.of(namespace="roundtrip_ns_2"))
        op2 = TransactionOperation.of(
            TransactionOperationType.CREATE, dest_metafile=ns2
        )
        txn.step(op2)

        txn.pause()

        # Step 3: Resume again, add third namespace, commit
        txn.resume()
        ns3 = Namespace.of(locator=NamespaceLocator.of(namespace="roundtrip_ns_3"))
        op3 = TransactionOperation.of(
            TransactionOperationType.CREATE, dest_metafile=ns3
        )
        txn.step(op3)

        # Final commit
        write_paths, success_log_path = txn.seal()

        assert txn.start_time < txn.end_time

        # Read and verify written namespaces
        for i, ns in enumerate([ns1, ns2, ns3]):
            written_path = write_paths[i]

            # Confirm file was created and is non-empty
            assert os.path.exists(written_path), f"Missing file: {written_path}"
            assert os.path.getsize(written_path) > 0, f"Empty file: {written_path}"

            # Deserialize and verify content
            deserialized_ns = Namespace.read(written_path)
            assert ns.equivalent_to(deserialized_ns), f"Namespace mismatch at index {i}"
            assert ns.locator.namespace == deserialized_ns.locator.namespace
            assert ns.locator_alias == deserialized_ns.locator_alias
            assert ns.properties == deserialized_ns.properties

        # Verify success log
        assert os.path.exists(success_log_path)
        assert success_log_path.endswith(str(txn.end_time))


class TestTransactionCommitMessage:
    """Test commit message preservation and retrieval for transactions."""

    def test_transaction_with_commit_message(self):
        """Test that commit messages are stored and retrievable from transactions."""
        commit_msg = "Test commit message for transaction functionality"

        # Create transaction with commit message
        txn = Transaction.of(commit_message=commit_msg)

        # Verify commit message is stored correctly
        assert txn.commit_message == commit_msg
        assert txn.get("commit_message") == commit_msg

    def test_transaction_without_commit_message(self):
        """Test that transactions work normally without commit messages."""
        # Create transaction without commit message
        txn = Transaction.of()

        # Verify no commit message is stored
        assert txn.commit_message is None
        assert txn.get("commit_message") is None

    def test_transaction_commit_message_setter(self):
        """Test that commit messages can be set after transaction creation."""
        # Create transaction without commit message
        txn = Transaction.of()
        assert txn.commit_message is None

        # Set commit message using property setter
        commit_msg = "Added commit message after creation"
        txn.commit_message = commit_msg

        # Verify commit message is stored correctly
        assert txn.commit_message == commit_msg
        assert txn.get("commit_message") == commit_msg

    def test_transaction_serialization_with_commit_message(self, temp_dir):
        """Test that commit messages persist through transaction serialization."""
        commit_msg = "Serialization test commit message"

        # Create namespace for testing
        ns = Namespace.of(locator=NamespaceLocator.of(namespace="serialization_test"))

        # Create transaction with commit message
        txn = Transaction.of(commit_message=commit_msg).start(temp_dir)
        op = TransactionOperation.of(TransactionOperationType.CREATE, dest_metafile=ns)
        txn.step(op)

        # Commit transaction (this should serialize the transaction with commit message)
        _, success_log_path = txn.seal()

        # Read the transaction log and verify commit message persisted
        txn_read = Transaction.read(success_log_path)
        assert txn_read.commit_message == commit_msg

        # Verify other transaction properties are intact
        assert txn_read.start_time == txn.start_time
        assert txn_read.end_time == txn.end_time
        assert len(txn_read.operations) == 1



from deltacat.storage import EntryParams


class TestEntryParams:
    def test_of_default(self):
        ep = EntryParams.of()
        assert ep == {}
        assert ep.equality_field_locators is None

    def test_of_with_equality_column_names(self):
        equality_column_names = ["col1", "col2"]
        ep = EntryParams.of(equality_field_locators=equality_column_names)
        assert ep == {"equality_field_locators": equality_column_names}
        assert ep.equality_field_locators == equality_column_names

    def test_equality_column_names_property(self):
        ep = EntryParams()
        assert ep.equality_field_locators is None
        equality_column_names = ["col1", "col2"]
        ep["equality_field_locators"] = equality_column_names
        assert ep.equality_field_locators == equality_column_names



import tempfile
from typing import Any, Dict, List, Optional, Set, Tuple, Callable
import uuid
import pytest

import pyarrow as pa
import ray
import pandas as pd

from deltacat.io.file_object_store import FileObjectStore
from pytest_benchmark.fixture import BenchmarkFixture

from deltacat.tests.compute.test_util_constant import (
    BASE_TEST_SOURCE_NAMESPACE,
    BASE_TEST_SOURCE_TABLE_NAME,
    BASE_TEST_SOURCE_TABLE_VERSION,
    DEFAULT_NUM_WORKERS,
    DEFAULT_WORKER_INSTANCE_CPUS,
)
from deltacat.compute.compactor.model.compactor_version import CompactorVersion
from deltacat.tests.compute.test_util_common import (
    create_src_w_deltas_destination_rebase_w_deltas_strategy_main,
    create_incremental_deltas_on_source_table_main,
    get_rci_from_partition,
    read_audit_file,
    PartitionKey,
    get_compacted_delta_locator_from_partition,
)
from deltacat.tests.compute.compact_partition_rebase_then_incremental_test_cases import (
    REBASE_THEN_INCREMENTAL_TEST_CASES,
)

from deltacat.types.media import StorageType
from deltacat.storage import (
    DeltaType,
    DeltaLocator,
    Partition,
    PartitionLocator,
    metastore,
)
from deltacat.types.media import ContentType
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.utils.placement import (
    PlacementGroupManager,
)
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)

"""
MODULE scoped fixtures
"""


@pytest.fixture(autouse=True, scope="module")
def setup_ray_cluster():
    ray.init(local_mode=True, ignore_reinit_error=True)
    yield
    ray.shutdown()


"""
FUNCTION scoped fixtures
"""


@pytest.fixture(autouse=True, scope="function")
def enable_bucketing_spec_validation(monkeypatch):
    """
    Enable the bucketing spec validation for all tests.
    This will help catch hash bucket drift in testing.
    """
    import deltacat.compute.compactor_v2.steps.merge

    monkeypatch.setattr(
        deltacat.compute.compactor_v2.steps.merge,
        "BUCKETING_SPEC_COMPLIANCE_PROFILE",
        "ASSERT",
    )


@pytest.mark.parametrize(
    [
        "test_name",
        "primary_keys",
        "sort_keys",
        "partition_keys_param",
        "partition_values_param",
        "input_deltas_param",
        "input_deltas_delta_type",
        "expected_terminal_compact_partition_result",
        "expected_terminal_exception",
        "expected_terminal_exception_message",
        "create_placement_group_param",
        "records_per_compacted_file_param",
        "hash_bucket_count_param",
        "read_kwargs_provider_param",
        "drop_duplicates_param",
        "skip_enabled_compact_partition_drivers",
        "assert_compaction_audit",
        "incremental_deltas",
        "rebase_expected_compact_partition_result",
        "compact_partition_func",
        "compactor_version",
    ],
    [
        (
            test_name,
            primary_keys,
            sort_keys,
            partition_keys_param,
            partition_values_param,
            input_deltas,
            input_deltas_delta_type,
            expected_terminal_compact_partition_result,
            expected_terminal_exception,
            expected_terminal_exception_message,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            drop_duplicates_param,
            read_kwargs_provider,
            skip_enabled_compact_partition_drivers,
            assert_compaction_audit,
            incremental_deltas,
            rebase_expected_compact_partition_result,
            compact_partition_func,
            compactor_version,
        )
        for test_name, (
            primary_keys,
            sort_keys,
            partition_keys_param,
            partition_values_param,
            input_deltas,
            input_deltas_delta_type,
            expected_terminal_compact_partition_result,
            expected_terminal_exception,
            expected_terminal_exception_message,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            drop_duplicates_param,
            read_kwargs_provider,
            skip_enabled_compact_partition_drivers,
            assert_compaction_audit,
            incremental_deltas,
            rebase_expected_compact_partition_result,
            compact_partition_func,
            compactor_version,
        ) in REBASE_THEN_INCREMENTAL_TEST_CASES.items()
    ],
    ids=[test_name for test_name in REBASE_THEN_INCREMENTAL_TEST_CASES],
)
def test_compact_partition_rebase_then_incremental_main(
    main_deltacat_storage_kwargs: Dict[str, Any],
    test_name: str,
    primary_keys: Set[str],
    sort_keys: List[Optional[Any]],
    partition_keys_param: Optional[List[PartitionKey]],
    partition_values_param: List[Optional[str]],
    input_deltas_param: List[pa.Array],
    input_deltas_delta_type: str,
    expected_terminal_compact_partition_result: pa.Table,
    expected_terminal_exception: BaseException,
    expected_terminal_exception_message: Optional[str],
    create_placement_group_param: bool,
    records_per_compacted_file_param: int,
    hash_bucket_count_param: int,
    drop_duplicates_param: bool,
    read_kwargs_provider_param: Any,
    incremental_deltas: List[Tuple[pa.Table, DeltaType, Optional[Dict[str, str]]]],
    rebase_expected_compact_partition_result: pa.Table,
    skip_enabled_compact_partition_drivers: List[CompactorVersion],
    assert_compaction_audit: Optional[Callable],
    compactor_version: Optional[CompactorVersion],
    compact_partition_func: Callable,
    benchmark: BenchmarkFixture,
):
    ds_mock_kwargs = main_deltacat_storage_kwargs
    """
    This test performs rebase compaction first, then incremental compaction on the same data.
    This tests the scenario where we first do a rebase (with different source/destination partitions)
    and then follow up with incremental compaction using the result of the rebase.

    This version uses the main metastore implementation instead of local storage.
    """

    """
    REBASE
    """
    partition_keys = partition_keys_param
    (
        source_table_stream,
        destination_table_stream,
        rebased_table_stream,
    ) = create_src_w_deltas_destination_rebase_w_deltas_strategy_main(
        sort_keys,
        partition_keys,
        input_deltas_param,
        input_deltas_delta_type,
        partition_values_param,
        ds_mock_kwargs,
    )

    # Convert partition values for partition lookup (same as in other helper functions)
    converted_partition_values_for_lookup = partition_values_param
    if partition_values_param and partition_keys:
        converted_partition_values_for_lookup = []
        for i, (value, pk) in enumerate(zip(partition_values_param, partition_keys)):
            if pk.key_type.value == "int":  # Use .value to get string representation
                converted_partition_values_for_lookup.append(int(value))
            elif pk.key_type.value == "timestamp":
                # Handle timestamp partition values
                if isinstance(value, str) and "T" in value and value.endswith("Z"):
                    ts = pd.to_datetime(value)
                    # Convert to microseconds since epoch for PyArrow timestamp[us]
                    converted_partition_values_for_lookup.append(
                        int(ts.timestamp() * 1_000_000)
                    )
                else:
                    converted_partition_values_for_lookup.append(value)
            else:
                converted_partition_values_for_lookup.append(value)

    source_partition: Partition = metastore.get_partition(
        source_table_stream.locator,
        converted_partition_values_for_lookup,
        **ds_mock_kwargs,
    )
    # Generate a destination partition ID based on the source partition
    destination_partition_id = str(uuid.uuid4())
    destination_partition_locator: PartitionLocator = PartitionLocator.of(
        destination_table_stream.locator,
        converted_partition_values_for_lookup,
        destination_partition_id,
    )
    all_column_names = metastore.get_table_version_column_names(
        destination_partition_locator.namespace,
        destination_partition_locator.table_name,
        destination_partition_locator.table_version,
        **ds_mock_kwargs,
    )
    rebased_partition: Partition = metastore.get_partition(
        rebased_table_stream.locator,
        converted_partition_values_for_lookup,
        **ds_mock_kwargs,
    )
    num_workers, worker_instance_cpu = DEFAULT_NUM_WORKERS, DEFAULT_WORKER_INSTANCE_CPUS
    total_cpus = num_workers * worker_instance_cpu
    pgm = None
    create_placement_group_param = False
    if create_placement_group_param:
        pgm = PlacementGroupManager(
            1, total_cpus, worker_instance_cpu, memory_per_bundle=4000000
        ).pgs[0]

    with tempfile.TemporaryDirectory() as test_dir:
        # Extract catalog from storage kwargs
        catalog = ds_mock_kwargs.get("inner")

        compact_partition_params = CompactPartitionParams.of(
            {
                "catalog": catalog,
                "compacted_file_content_type": ContentType.PARQUET,
                "dd_max_parallelism_ratio": 1.0,
                "deltacat_storage": metastore,
                "deltacat_storage_kwargs": ds_mock_kwargs,
                "destination_partition_locator": destination_partition_locator,
                "hash_bucket_count": hash_bucket_count_param,
                "last_stream_position_to_compact": source_partition.stream_position,
                "list_deltas_kwargs": {
                    **ds_mock_kwargs,
                    **{"equivalent_table_types": []},
                },
                "object_store": FileObjectStore(test_dir),
                "original_fields": {
                    "pk_col_1",
                    "pk_col_2",
                    "sk_col_1",
                    "sk_col_2",
                    "col_1",
                    "col_2",
                    "region_id",
                },
                "pg_config": pgm,
                "primary_keys": primary_keys,
                "all_column_names": all_column_names,
                "read_kwargs_provider": read_kwargs_provider_param,
                "rebase_source_partition_locator": source_partition.locator,
                "records_per_compacted_file": records_per_compacted_file_param,
                "source_partition_locator": rebased_partition.locator,
                "sort_keys": sort_keys if sort_keys else None,
            }
        )
        # execute
        benchmark(compact_partition_func, compact_partition_params)
        compacted_delta_locator: DeltaLocator = (
            get_compacted_delta_locator_from_partition(
                destination_partition_locator,
                metastore,
                catalog=catalog,
            )
        )
        tables = metastore.download_delta(
            compacted_delta_locator,
            storage_type=StorageType.LOCAL,
            **ds_mock_kwargs,
        )
        actual_rebase_compacted_table = pa.concat_tables(tables)
        all_column_names = metastore.get_table_version_column_names(
            destination_partition_locator.namespace,
            destination_partition_locator.table_name,
            destination_partition_locator.table_version,
            **ds_mock_kwargs,
        )
        # if no primary key is specified then sort by sort_key for consistent assertion
        sorting_cols: List[Any] = (
            [(val, "ascending") for val in primary_keys]
            if primary_keys
            else [pa_key for key in sort_keys for pa_key in key.arrow]
            if sort_keys
            else []
        )
        rebase_expected_compact_partition_result = (
            rebase_expected_compact_partition_result.combine_chunks().sort_by(
                sorting_cols
            )
        )
        actual_rebase_compacted_table = (
            actual_rebase_compacted_table.combine_chunks().sort_by(sorting_cols)
        )
        assert actual_rebase_compacted_table.equals(
            rebase_expected_compact_partition_result
        ), f"{actual_rebase_compacted_table} does not match {rebase_expected_compact_partition_result}"

        """
        INCREMENTAL
        """
        (
            source_partition_locator_w_deltas,
            new_delta,
            incremental_delta_length,
            has_delete_deltas,
        ) = create_incremental_deltas_on_source_table_main(
            BASE_TEST_SOURCE_NAMESPACE,
            BASE_TEST_SOURCE_TABLE_NAME,
            BASE_TEST_SOURCE_TABLE_VERSION,
            source_table_stream,
            partition_values_param,
            incremental_deltas,
            ds_mock_kwargs,
        )

        # Handle empty incremental deltas case
        if new_delta is None:
            # For empty incremental deltas, the expected result should be the same as rebase result
            # Skip incremental compaction and just verify the rebase result
            actual_compact_partition_result = actual_rebase_compacted_table
            compaction_audit = None
        else:
            # Perform incremental compaction when there are actual deltas
            last_stream_position = new_delta.stream_position

            compact_partition_params = CompactPartitionParams.of(
                {
                    "catalog": catalog,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": metastore,
                    "deltacat_storage_kwargs": ds_mock_kwargs,
                    "destination_partition_locator": compacted_delta_locator.partition_locator,
                    "drop_duplicates": drop_duplicates_param,
                    "hash_bucket_count": hash_bucket_count_param,
                    "last_stream_position_to_compact": last_stream_position,
                    "list_deltas_kwargs": {
                        **ds_mock_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "object_store": FileObjectStore(test_dir),
                    "original_fields": {
                        "pk_col_1",
                        "pk_col_2",
                        "sk_col_1",
                        "sk_col_2",
                        "col_1",
                        "col_2",
                        "region_id",
                    },
                    "pg_config": pgm,
                    "primary_keys": primary_keys,
                    "all_column_names": all_column_names,
                    "read_kwargs_provider": read_kwargs_provider_param,
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": records_per_compacted_file_param,
                    "source_partition_locator": source_partition_locator_w_deltas,
                    "sort_keys": sort_keys if sort_keys else None,
                }
            )
            if expected_terminal_exception:
                with pytest.raises(expected_terminal_exception) as exc_info:
                    compact_partition_func(compact_partition_params)
                assert expected_terminal_exception_message in str(exc_info.value)
                return
            compact_partition_func(compact_partition_params)
            # assert
            compacted_delta_locator: DeltaLocator = (
                get_compacted_delta_locator_from_partition(
                    destination_partition_locator, metastore, catalog=catalog
                )
            )
            tables = metastore.download_delta(
                compacted_delta_locator,
                storage_type=StorageType.LOCAL,
                **ds_mock_kwargs,
            )
            actual_compact_partition_result = pa.concat_tables(tables)

            # Get compaction audit for verification if needed
            round_completion_info = get_rci_from_partition(
                destination_partition_locator, metastore, catalog=catalog
            )
            # Get catalog root for audit file resolution
            catalog_root = catalog.root

            compaction_audit_obj: dict = read_audit_file(
                round_completion_info.compaction_audit_url, catalog_root
            )
            compaction_audit = CompactionSessionAuditInfo(**compaction_audit_obj)

        # Verify the final result
        actual_compact_partition_result = (
            actual_compact_partition_result.combine_chunks().sort_by(sorting_cols)
        )
        expected_terminal_compact_partition_result = (
            expected_terminal_compact_partition_result.combine_chunks().sort_by(
                sorting_cols
            )
        )
        assert actual_compact_partition_result.equals(
            expected_terminal_compact_partition_result
        ), f"{actual_compact_partition_result} does not match {expected_terminal_compact_partition_result}"

        if assert_compaction_audit is not None and compaction_audit is not None:
            if not assert_compaction_audit(compactor_version, compaction_audit):
                pytest.fail("Compaction audit assertion failed")
        return
